from urllib.parse import quote

from IPython.display import HTML  # type: ignore

from dark.fastq import FastqReads


def NCBISequenceLinkURL(title, field=None, delim="|"):
    """
    Given a sequence title, like
        "acc|GENBANK|AY516849.1|GENBANK|42768646 Homo sapiens",
    return the URL of a link to the info page at NCBI.

    @param title: The C{str} sequence title to produce a link URL for.
    @param field: The C{int} field number to use (as delimited by C{delim})
        or C{None} if no field splitting should be done.
    @param delim: The C{str} to split the title on (if C{field} is not
        C{None}).
    @return: A C{str} URL.
    """
    if field is None:
        ref = title
    else:
        try:
            ref = title.split(delim)[field]
        except IndexError:
            raise IndexError(
                "Could not extract field %d from sequence title %r" % (field, title)
            )
    return "http://www.ncbi.nlm.nih.gov/nuccore/" + quote(ref)


def NCBISequenceLink(title, field=None, delim="|"):
    """
    Given a sequence title, like
        "acc|GENBANK|AY516849.1|GENBANK|42768646 Homo sapiens",
    return an HTML <A> tag displaying a link to the info page at NCBI.

    @param title: The C{str} sequence title to produce a link URL for.
    @param field: The C{int} field number to use (as delimited by C{delim})
        or C{None} if no field splitting should be done.
    @param delim: The C{str} to split the title on (if C{field} is not
        C{None}).
    @return: A C{str} HTML <A> tag.
    """
    return '<a href="%s" target="_blank">%s</a>' % (
        NCBISequenceLinkURL(title, field, delim),
        title,
    )


def _sortHTML(titlesAlignments, by, limit=None):
    """
    Return an C{IPython.display.HTML} object with the alignments sorted by the
    given attribute.

    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    @param by: A C{str}, one of 'length', 'maxScore', 'medianScore',
        'readCount', or 'title'.
    @param limit: An C{int} limit on the number of results to show.
    @return: An HTML instance with sorted titles and information about
        hit read count, length, and e-values.
    """
    out = []
    for i, title in enumerate(titlesAlignments.sortTitles(by), start=1):
        if limit is not None and i > limit:
            break
        titleAlignments = titlesAlignments[title]
        link = NCBISequenceLink(title, title)
        out.append(
            "%3d: reads=%d, len=%d, max=%s median=%s<br/>"
            "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%s"
            % (
                i,
                titleAlignments.readCount(),
                titleAlignments.subjectLength,
                titleAlignments.bestHsp().score.score,
                titleAlignments.medianScore(),
                link,
            )
        )
    return HTML("<pre>" + "<br/>".join(out) + "</pre>")


def summarizeTitlesByTitle(titlesAlignments, limit=None):
    """
    Sort match titles by title

    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    @param limit: An C{int} limit on the number of results to show.
    @return: An C{IPython.display.HTML} instance with match titles sorted by
        title.
    """
    return _sortHTML(titlesAlignments, "title", limit)


def summarizeTitlesByCount(titlesAlignments, limit=None):
    """
    Sort match titles by read count.

    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    @param limit: An C{int} limit on the number of results to show.
    @return: An C{IPython.display.HTML} instance with match titles sorted by
        read count.
    """
    return _sortHTML(titlesAlignments, "readCount", limit)


def summarizeTitlesByLength(titlesAlignments, limit=None):
    """
    Sort match titles by sequence length.

    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    @param limit: An C{int} limit on the number of results to show.
    @return: An C{IPython.display.HTML} instance with match titles sorted by
        sequence length.
    """
    return _sortHTML(titlesAlignments, "length", limit)


def summarizeTitlesByMaxScore(titlesAlignments, limit=None):
    """
    Sort hit titles by maximum score.

    @param titlesAlignments: A L{dark.blast.BlastMatchs} instance.
    @param limit: An C{int} limit on the number of results to show.
    @return: An C{IPython.display.HTML} instance with hit titles sorted by
        max score.
    """
    return _sortHTML(titlesAlignments, "maxScore", limit)


def summarizeTitlesByMedianScore(titlesAlignments, limit=None):
    """
    Sort match titles by median score.

    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    @param limit: An C{int} limit on the number of results to show.
    @return: An C{IPython.display.HTML} instance with match titles sorted by
        median score.
    """
    return _sortHTML(titlesAlignments, "medianScore", limit)


class AlignmentPanelHTMLWriter:
    """
    Produces HTML details of a rectangular panel of graphs that each
    contain an alignment graph against a given sequence. This is
    supplementary output info for the AlignmentPanel class in graphics.py.

    @param outputDir: The C{str} directory to write files into.
    @param titlesAlignments: A L{dark.titles.TitlesAlignments} instance.
    """

    def __init__(self, outputDir, titlesAlignments):
        self._outputDir = outputDir
        self._titlesAlignments = titlesAlignments
        self._images = []

    def addImage(self, imageBasename, title, graphInfo):
        self._images.append(
            {"graphInfo": graphInfo, "imageBasename": imageBasename, "title": title}
        )

    def close(self):
        with open("%s/index.html" % self._outputDir, "w") as fp:
            self._writeHeader(fp)
            self._writeBody(fp)
            self._writeFooter(fp)
        with open("%s/style.css" % self._outputDir, "w") as fp:
            self._writeCSS(fp)

    def _writeHeader(self, fp):
        fp.write(
            """\
<html>
  <head>
    <title>Read alignments for %d matched subjects</title>
    <link rel="stylesheet" type="text/css" href="style.css">
  </head>
  <body>
    <div id="content">
        """
            % len(self._images)
        )

    def _writeBody(self, fp):
        fp.write(
            "<h1>Read alignments for %d matched subjects</h1>\n" % len(self._images)
        )

        # Write out an alignment panel as a table.
        cols = 6
        fp.write("<table><tbody>\n")

        for i, image in enumerate(self._images):
            title = image["title"]

            if i % cols == 0:
                fp.write("<tr>\n")

            fp.write(
                '<td><a id="small_%d"></a><a href="#big_%d"><img src="%s" '
                'class="thumbnail"/></a></td>\n' % (i, i, image["imageBasename"])
            )

            if i % cols == cols - 1:
                fp.write("</tr>")

        # Add empty cells to the final table row, and close the row, if
        # necessary.
        if i % cols < cols - 1:
            while i % cols < cols - 1:
                fp.write("<td>&nbsp;</td>\n")
                i += 1
            fp.write("</tr>\n")

        fp.write("</tbody></table>\n")

        # Write out the full images with additional detail.
        for i, image in enumerate(self._images):
            title = image["title"]
            titleAlignments = self._titlesAlignments[title]
            graphInfo = image["graphInfo"]
            readFormat = self._writeFASTA(i, image)
            fp.write(
                """
      <a id="big_%d"></a>
      <h3>%d: %s</h3>
      <p>
            Length: %d.
            Read count: %d.
            HSP count: %d.
            <a href="%d.%s">%s</a>.
            <a href="#small_%d">Top panel.</a>
"""
                % (
                    i,
                    i,
                    title,
                    titleAlignments.subjectLength,
                    titleAlignments.readCount(),
                    titleAlignments.hspCount(),
                    i,
                    readFormat,
                    readFormat,
                    i,
                )
            )

            url = NCBISequenceLinkURL(title)
            if url:
                fp.write('<a href="%s" target="_blank">NCBI</a>.' % url)

            # Write out feature information.
            if graphInfo["features"] is None:
                # Feature lookup was False (or we were offline).
                pass
            elif len(graphInfo["features"]) == 0:
                fp.write("There were no features.")
            else:
                fp.write('<a href="%s">Features</a>' % self._writeFeatures(i, image))

            # Write out the titles that this title invalidated due to its
            # read set.
            readSetFilter = self._titlesAlignments.readSetFilter
            if readSetFilter:
                invalidated = readSetFilter.invalidates(title)
                if invalidated:
                    nInvalidated = len(invalidated)
                    fp.write(
                        "<br/>This title invalidated %d other%s due to "
                        "its read set:<ul>"
                        % (nInvalidated, "" if nInvalidated == 1 else "s")
                    )
                    for title in invalidated:
                        fp.write("<li>%s</li>" % title)
                    fp.write("</ul>")

            fp.write('</p><img src="%s" class="full-size"/>' % image["imageBasename"])

    def _writeFooter(self, fp):
        fp.write(
            """\
    </div>
  </body>
</html>
"""
        )

    def _writeCSS(self, fp):
        fp.write(
            """\
#content {
  width: 95%;
  margin: auto;
}
img.thumbnail {
  height: 300px;
}
img.full-size {
  height: 900px;
}
"""
        )

    def _writeFASTA(self, i, image):
        """
        Write a FASTA file containing the set of reads that hit a sequence.

        @param i: The number of the image in self._images.
        @param image: A member of self._images.
        @return: A C{str}, either 'fasta' or 'fastq' indicating the format
            of the reads in C{self._titlesAlignments}.
        """
        if isinstance(self._titlesAlignments.readsAlignments.reads, FastqReads):
            format_ = "fastq"
        else:
            format_ = "fasta"
        filename = "%s/%d.%s" % (self._outputDir, i, format_)
        titleAlignments = self._titlesAlignments[image["title"]]
        with open(filename, "w") as fp:
            for titleAlignment in titleAlignments:
                fp.write(titleAlignment.read.toString(format_))
        return format_

    def _writeFeatures(self, i, image):
        """
        Write a text file containing the features as a table.

        @param i: The number of the image in self._images.
        @param image: A member of self._images.
        @return: The C{str} features file name - just the base name, not
            including the path to the file.
        """
        basename = "features-%d.txt" % i
        filename = "%s/%s" % (self._outputDir, basename)
        featureList = image["graphInfo"]["features"]
        with open(filename, "w") as fp:
            for feature in featureList:
                fp.write("%s\n\n" % feature.feature)
        return basename


def readCountText(readCountColors, count, linkText=None):
    """
    Produce colored read count text.

    @param readCountColors: Either a C{dark.colors.colorsForCounts}
        instance or C{None} for no read count coloring.
    @param count: An C{int} read count.
    @param linkText: A C{str} for the HTML link text. If C{None}, the
        count will be used.
    @return: An HTML span C{str} colored according to the read count,
        or just the string of the count if no color information is given.
    """
    if readCountColors:
        _class = readCountColors.thresholdToCssName(
            readCountColors.thresholdForCount(count)
        )
        return f'<span class="{_class}">{count}</span>'
    else:
        return linkText or str(count)



import sqlite3
from pathlib import Path


def sqliteConnect(filename: str | Path) -> sqlite3.Connection:
    """
    Connect to sqlite3, with a more informative error in case of failure.

    @param : A C{str} or C{Path} database filename.
    @return: An sqlite3 connection.
    """
    try:
        return sqlite3.connect(filename)
    except sqlite3.OperationalError as e:
        # The current (2024-06-01) sqlite3.OperationalError does not
        # include the filename that sqlite3 was trying to open, which
        # is not very helpful. Raise an error that does.
        raise sqlite3.OperationalError(
            f"Could not open sqlite3 database file {str(filename)!r}: {e}"
        )



#!/usr/bin/env python

import argparse
import csv
import sys

from summarize_ranges.ranges import RangeSummarizer

from dark.dna import AMBIGUOUS
from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


def formatRanges(rs: RangeSummarizer, highAdjust: int) -> str:
    result = []
    for low, high in rs.ranges():
        high -= highAdjust
        if low == high:
            result.append(str(low))
        else:
            result.append(f"{low}-{high}")

    return ",".join(result)


def main():
    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on stdin, write information about the "
            "counts of characters in the sequences to stdout."
        )
    )

    parser.add_argument(
        "--ignoreCase",
        "-i",
        action="store_true",
        help="Map all characters to uppercase.",
    )

    parser.add_argument(
        "--sep", default="\t", help="The output field separator character."
    )

    parser.add_argument(
        "--omitIds",
        action="store_true",
        help="Do not print sequence ids as the first field.",
    )

    parser.add_argument(
        "--omitHeader",
        dest="includeHeader",
        action="store_false",
        help="Do not print a header.",
    )

    parser.add_argument(
        "--ambiguous",
        action="store_true",
        help="If --chars is not given, print counts for all ambiguous DNA codes.",
    )

    parser.add_argument(
        "--omitZeroes",
        action="store_true",
        help="Do not print chars whose counts are zero.",
    )

    parser.add_argument(
        "--locations",
        action="store_true",
        help="Also print the 1-based genome locations of the matched characters.",
    )

    parser.add_argument(
        "--sort",
        action="store_true",
        help=(
            "Sort the characters for each sequence by decreasing count (with draws "
            "broken in favour of alphabetically increasing character)."
        ),
    )

    parser.add_argument(
        "--python",
        action="store_true",
        help=(
            "Print 0-based Python closed-open location ranges (i.e., the upper bound "
            "is not included) when --locations is used."
        ),
    )

    parser.add_argument(
        "--chars",
        "-c",
        help="The characters to count. Default is all IUPAC nucleotide codes.",
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()

    omitZeroes = args.omitZeroes
    addLocations = args.locations
    highAdjust = 0 if args.python else 1
    omitIds = args.omitIds
    reads = parseFASTACommandLineOptions(args)
    writerow = csv.writer(sys.stdout, delimiter=args.sep).writerow

    if args.chars:
        chars = args.chars
    else:
        if args.ambiguous:
            chars = "".join(
                sorted(c for (c, codons) in AMBIGUOUS.items() if len(codons) > 1)
            )
        else:
            chars = "".join(sorted(AMBIGUOUS))

    if args.ignoreCase:
        chars = chars.upper()

    if args.includeHeader:
        header = [] if omitIds else ["ID"]
        header.extend(("Char", "Count", "Length", "Fraction"))
        if addLocations:
            header.append("Location(s)")
        writerow(header)

    def key(row):
        return (row[1], row[0]) if omitIds else (row[2], row[1])

    for read in reads:
        length = len(read)
        sequence = read.sequence.upper() if args.ignoreCase else read.sequence
        rows = []
        for char in chars:
            if addLocations:
                locations = []
                rs = RangeSummarizer()

            count = 0
            for i, c in enumerate(sequence):
                if c == char:
                    count += 1
                    if addLocations:
                        rs.add(str(i + 1))
                        locations.append(i + 1)

            if count or not omitZeroes:
                row = [] if omitIds else [read.id]
                row.extend([char, count, length, f"{count / length:.4f}"])
                if addLocations:
                    row.append(formatRanges(rs, highAdjust))
                rows.append(row)

        if args.sort:
            rows.sort(key=key, reverse=True)

        for row in rows:
            writerow(row)


if __name__ == "__main__":
    main()



import os
from typing import Optional, Tuple

import matplotlib
import numpy as np

if not os.environ.get("DISPLAY"):
    # Use non-interactive Agg backend
    matplotlib.use("Agg")
from matplotlib import pyplot as plt

from dark.entrez import getSequence


class Feature:
    """
    An offset-adjusted feature, with start and stop attributes and methods to
    return a textual description and a legend label.

    @param feature: A BioPython feature.
    @param subfeature: A C{bool} to indicate if a feature is actually a
        subfeature.
    """

    def __init__(self, feature, subfeature=False):
        self.feature = feature
        self.color = None  # Should be set with setColor
        self.subfeature = subfeature
        self.start = int(feature.location.start)
        self.end = int(feature.location.end)

    def setColor(self, color):
        """
        An explicit method to set a feature's (plotting) color.

        @param color: A C{str} color.
        """
        self.color = color

    def legendLabel(self):
        """
        Provide a textual description of the feature and its qualifiers to be
        used as a label in a plot legend.

        @return: A C{str} description of the feature.
        """
        excludedQualifiers = set(
            (
                "codon_start",
                "db_xref",
                "protein_id",
                "region_name",
                "ribosomal_slippage",
                "rpt_type",
                "translation",
                "transl_except",
                "transl_table",
            )
        )
        maxValueLength = 30
        result = []
        if self.feature.qualifiers:
            for qualifier in sorted(self.feature.qualifiers):
                if qualifier not in excludedQualifiers:
                    value = ", ".join(self.feature.qualifiers[qualifier])
                    if qualifier == "site_type" and value == "other":
                        continue
                    if len(value) > maxValueLength:
                        value = value[: maxValueLength - 3] + "..."
                    result.append("%s: %s" % (qualifier, value))
        return "%d-%d %s%s.%s" % (
            int(self.feature.location.start),
            int(self.feature.location.end),
            self.feature.type,
            " (subfeature)" if self.subfeature else "",
            " " + ", ".join(result) if result else "",
        )


class FeatureList(list):
    """
    Provide access to a list of L{Feature} objects.

    @param title: A C{str} sequence title from a BLAST hit. Of the form
        'gi|63148399|gb|DQ011818.1| Description...'.
    @param database: The S{str} name of the Entrez database to search.
    @param wantedTypes: A C{tuple} of feature types that are of interest.
        Feature whose types are not in this list will be ignored.
    @param sequenceFetcher: A function that takes a sequence title and a
        database name and returns a C{Bio.SeqIO} instance. If C{None}, use
        L{dark.entrez.getSequence}.
    """

    def __init__(self, title, database, wantedTypes, sequenceFetcher=None):
        list.__init__(self)
        self.offline = False
        sequenceFetcher = sequenceFetcher or getSequence
        try:
            record = sequenceFetcher(title, db=database)
        except ValueError:
            # Ignore. See https://github.com/acorg/dark-matter/issues/124
            return
        if record is None:
            self.offline = True
        else:
            wantedTypes = set(wantedTypes)
            for feature in record.features:
                if feature.type in wantedTypes:
                    self.append(Feature(feature))

            # Assign colors to features.
            colormap = plt.cm.coolwarm
            colors = [colormap(i) for i in np.linspace(0.0, 0.99, len(self))]
            for feature, color in zip(self, colors):
                feature.setColor(color)


class _FeatureAdder:
    """
    Look up features for a title, and provide a method to add them to a figure
    as well as returning them.
    """

    TITLE_FONTSIZE = 16
    FONTSIZE = 20
    MAX_FEATURES_TO_DISPLAY = 50
    DATABASE = Optional[str]  # Set in subclasses.
    WANTED_TYPES = Optional[Tuple[str, ...]]  # Set in subclasses.

    def __init__(self):
        self.tooManyFeaturesToPlot = False

    def add(self, fig, title, minX, maxX, offsetAdjuster=None, sequenceFetcher=None):
        """
        Find the features for a sequence title. If there aren't too many, add
        the features to C{fig}. Return information about the features, as
        described below.

        @param fig: A matplotlib figure.
        @param title: A C{str} sequence title from a BLAST hit. Of the form
            'gi|63148399|gb|DQ011818.1| Description...'.
        @param minX: The smallest x coordinate.
        @param maxX: The largest x coordinate.
         @param offsetAdjuster: a function for adjusting feature X axis offsets
            for plotting.
        @param sequenceFetcher: A function that takes a sequence title and a
            database name and returns a C{Bio.SeqIO} instance. If C{None}, use
            L{dark.entrez.getSequence}.
        @return: If we seem to be offline, return C{None}. Otherwise, return
            a L{FeatureList} instance.
        """

        offsetAdjuster = offsetAdjuster or (lambda x: x)

        fig.set_title("Target sequence features", fontsize=self.TITLE_FONTSIZE)
        fig.set_yticks([])

        features = FeatureList(
            title, self.DATABASE, self.WANTED_TYPES, sequenceFetcher=sequenceFetcher
        )

        if features.offline:
            fig.text(
                minX + (maxX - minX) / 3.0,
                0,
                "You (or Genbank) appear to be offline.",
                fontsize=self.FONTSIZE,
            )
            fig.axis([minX, maxX, -1, 1])
            return None

        # If no interesting features were found, display a message saying
        # so in the figure.  Otherwise, if we don't have too many features
        # to plot, add the feature info to the figure.
        nFeatures = len(features)
        if nFeatures == 0:
            # fig.text(minX + (maxX - minX) / 3.0, 0, 'No features found',
            #          fontsize=self.FONTSIZE)
            fig.text(
                0.5,
                0.5,
                "No features found",
                horizontalalignment="center",
                verticalalignment="center",
                transform=fig.transAxes,
                fontsize=self.FONTSIZE,
            )
            fig.axis([minX, maxX, -1, 1])
        elif nFeatures <= self.MAX_FEATURES_TO_DISPLAY:
            # Call the method in our subclass to do the figure display.
            self._displayFeatures(fig, features, minX, maxX, offsetAdjuster)
        else:
            self.tooManyFeaturesToPlot = True
            # fig.text(minX + (maxX - minX) / 3.0, 0,
            # 'Too many features to plot.', fontsize=self.FONTSIZE)
            fig.text(
                0.5,
                0.5,
                "Too many features to plot",
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=self.FONTSIZE,
                transform=fig.transAxes,
            )
            fig.axis([minX, maxX, -1, 1])

        return features

    def _displayFeatures(self, fig, features, minX, maxX, offsetAdjuster):
        """
        Add the given C{features} to the figure in C{fig}.

        @param fig: A matplotlib figure.
        @param features: A C{FeatureList} instance.
        @param minX: The smallest x coordinate.
        @param maxX: The largest x coordinate.
        @param offsetAdjuster: a function for adjusting feature X axis offsets
            for plotting.
        """
        raise NotImplementedError("_displayFeatures must be implemented in a subclass.")


class ProteinFeatureAdder(_FeatureAdder):
    """
    Subclass L{_FeatureAdder} with a method to add protein features to a
    figure.
    """

    DATABASE = "protein"
    WANTED_TYPES = ("CDS", "mat_peptide", "rRNA", "Site", "Region")

    def _displayFeatures(self, fig, features, minX, maxX, offsetAdjuster):
        """
        Add the given C{features} to the figure in C{fig}.

        @param fig: A matplotlib figure.
        @param features: A C{FeatureList} instance.
        @param minX: The smallest x coordinate.
        @param maxX: The largest x coordinate.
        @param offsetAdjuster: a function for adjusting feature X axis offsets
            for plotting.
        """
        labels = []
        for index, feature in enumerate(features):
            fig.plot(
                [offsetAdjuster(feature.start), offsetAdjuster(feature.end)],
                [index * -0.2, index * -0.2],
                color=feature.color,
                linewidth=2,
            )
            labels.append(feature.legendLabel())

        # Note that minX and maxX do not need to be adjusted by the offset
        # adjuster. They are the already-adjusted min/max values as
        # computed in computePlotInfo in blast.py
        fig.axis([minX, maxX, (len(features) + 1) * -0.2, 0.2])

        if labels:
            # Put a legend above the figure.
            box = fig.get_position()
            fig.set_position([box.x0, box.y0, box.width, box.height * 0.2])
            fig.legend(
                labels,
                loc="lower center",
                bbox_to_anchor=(0.5, 1.4),
                fancybox=True,
                shadow=True,
                ncol=2,
            )


class NucleotideFeatureAdder(_FeatureAdder):
    """
    Subclass L{_FeatureAdder} with a method to add nucleotide features to a
    figure.
    """

    DATABASE = "nucleotide"
    WANTED_TYPES = (
        "CDS",
        "LTR",
        "mat_peptide",
        "misc_feature",
        "misc_structure",
        "repeat_region",
        "rRNA",
    )

    def _displayFeatures(self, fig, features, minX, maxX, offsetAdjuster):
        """
        Add the given C{features} to the figure in C{fig}.

        @param fig: A matplotlib figure.
        @param features: A C{FeatureList} instance.
        @param minX: The smallest x coordinate.
        @param maxX: The largest x coordinate.
        @param offsetAdjuster: a function for adjusting feature X axis offsets
            for plotting.
        """
        frame = None
        labels = []
        for feature in features:
            start = offsetAdjuster(feature.start)
            end = offsetAdjuster(feature.end)
            if feature.subfeature:
                subfeatureFrame = start % 3
                if subfeatureFrame == frame:
                    # Move overlapping subfeatures down a little to make them
                    # visible.
                    y = subfeatureFrame - 0.2
                else:
                    y = subfeatureFrame
            else:
                frame = start % 3
                # If we have a polyprotein, shift it up slightly so we can see
                # its components below it.
                product = feature.feature.qualifiers.get("product", [""])[0]
                if product.lower().find("polyprotein") > -1:
                    y = frame + 0.2
                else:
                    y = frame
            fig.plot([start, end], [y, y], color=feature.color, linewidth=2)
            labels.append(feature.legendLabel())

        # Note that minX and maxX do not need to be adjusted by the offset
        # adjuster. They are the already-adjusted min/max values as
        # computed in computePlotInfo in blast.py
        fig.axis([minX, maxX, -0.5, 2.5])
        fig.set_yticks(np.arange(3))
        fig.set_ylabel("Frame")

        if labels:
            # Put a legend above the figure.
            box = fig.get_position()
            fig.set_position([box.x0, box.y0, box.width, box.height * 0.3])
            fig.legend(
                labels,
                loc="lower center",
                bbox_to_anchor=(0.5, 2.5),
                fancybox=True,
                shadow=True,
                ncol=2,
            )



#!/usr/bin/env python

import argparse
import sys
from collections import Counter, defaultdict
from itertools import chain, cycle
from operator import itemgetter

import matplotlib.pyplot as plt
import plotly.express as px
import plotly.graph_objs as go
from plotly.io import write_html, write_image

from dark.reads import DNARead, addFASTACommandLineOptions, parseFASTACommandLineOptions
from dark.windowedIdentity import WindowedIdentity, addCommandLineOptions


def makeParser() -> argparse.ArgumentParser:
    """
    Make an argument parser.
    """

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Make a plot showing windowed nucleotide identity across an alignment."
        ),
    )

    addFASTACommandLineOptions(parser)
    addCommandLineOptions(parser)

    parser.add_argument(
        "--html",
        help="The name of an HTML file to write a plotly image to.",
    )

    parser.add_argument(
        "--out",
        help=(
            "The name of the file to write the image to. The file suffix "
            "will determine the output format. The image will be created with "
            "plotly unless --matplotlib is specified."
        ),
    )

    parser.add_argument(
        "--title",
        help="The plot title.",
    )

    parser.add_argument(
        "--legendTitle",
        default="Sample",
        help="The legend title.",
    )

    parser.add_argument(
        "--dryRun",
        "-n",
        action="store_true",
        help="Just print what would be compared without actually doing it.",
    )

    parser.add_argument(
        "--bestOnly",
        action="store_true",
        help="Only show the best matching sequences for each window.",
    )

    parser.add_argument(
        "--matplotlib",
        action="store_true",
        help="Use matplotlib to make a static image. Otherwise, plotly will be used.",
    )

    parser.add_argument(
        "--y01",
        action="store_true",
        help="Set the y-axis limits to (0.0, 1.0). Only applicable for matplotlib plots.",  # noqa: E501
    )

    parser.add_argument(
        "--ambiguous",
        dest="strict",
        action="store_false",
        help=(
            "Count compatible ambiguous nucleotide codes as matching when computing "
            "identity."
        ),
    )

    parser.add_argument(
        "--markerSize",
        type=int,
        help="The size of the markers in plot lines.",
    )

    return parser


def plotMatplotlib(
    centers: list[float],
    identity: dict[DNARead, list[float]],
    title: str,
    args: argparse.Namespace,
) -> None:
    """
    Make a matplotlib image.

    @param centers: The x-axis centers of the genome windows.
    @param identity: A mapping from sequences (that have been matched against the
        reference) to a list of float identities (one for each window).
    @param title: The title of the plot.
    @param args: The command-line arguments.
    """
    figure, ax = plt.subplots(1, 1, figsize=(10, 8))

    for read, identities in identity.items():
        ax.plot(centers, identities, label=f"{read.id}")

    if args.y01:
        ax.set_ylim((0.0, 1.0))

    ax.grid(axis="x", color="0.95")
    ax.legend(title=args.legendTitle)
    figure.suptitle(title)

    figure.savefig(args.out)


def getBestOnly(
    centers: list[float], identity: dict[DNARead, list[float]], args: argparse.Namespace
) -> list[go.Scatter]:
    """
    Find the best-matching sequence(s) for each window and return scatter plots for them.

    @param centers: The x-axis centers of the genome windows.
    @param identity: A mapping from sequences (that have been matched against the
        reference) to a list of float identities (one for each window).
    @param title: The title of the plot.
    @param args: The command-line arguments.
    """  # noqa: E501
    tolerance = 0.99
    delta = 0.005
    half = args.jump / 2.0
    inLegend = set()

    mode, marker = (
        ("lines+markers", dict(size=args.markerSize))
        if args.markerSize  # i.e., not 0 and not None.
        else ("lines", None)
    )

    # Keep track of how many times each read has a window that is amongst the
    # best. We'll use that to update the legend to display the count.
    bestCounts: Counter[str] = Counter()

    # scatters will hold a list of go.Scatter instances, keyed by readId.
    scatters = defaultdict(list)

    colors: dict[str, str] = {}
    colorIterator = cycle(
        chain(px.colors.qualitative.Dark2, px.colors.qualitative.Pastel2)
    )

    for jumpIndex in range(len(centers)):
        windowIdentities = []
        # It's less efficient to loop over .items() repeatedly like this, but I think it
        # makes for simpler code since we process each window in turn.
        for read, identities in identity.items():
            windowIdentities.append((identities[jumpIndex], read.id))
        # Sort first by ascending read id.
        windowIdentities.sort(key=itemgetter(1))
        # Then by descending identity.
        windowIdentities.sort(key=itemgetter(0), reverse=True)

        bestMatches = [windowIdentities[0]]
        bestIdentity = bestMatches[0][0]

        # Collect all sequences that are within the tolerance of the best identity.
        for thisIdentity, readId in windowIdentities[1:]:
            if thisIdentity >= tolerance * bestIdentity:
                bestMatches.append((thisIdentity, readId))
            else:
                break

        center = centers[jumpIndex]
        x = [center - half, center + half]

        # Collect the best matches.
        lastIdentity = 2.0  # Must be > 1.0 to start things off.
        for thisIdentity, readId in bestMatches:
            bestCounts[readId] += 1
            showlegend = readId not in inLegend
            inLegend.add(readId)
            if thisIdentity >= lastIdentity:
                adjustedIdentity = lastIdentity - delta
            else:
                adjustedIdentity = thisIdentity
            lastIdentity = adjustedIdentity
            try:
                color = colors[readId]
            except KeyError:
                color = colors[readId] = next(colorIterator)
            scatters[readId].append(
                go.Scatter(
                    x=x,
                    y=[adjustedIdentity, adjustedIdentity],
                    name=readId,
                    text=f"{readId} {int(x[0])}-{int(x[1])} {thisIdentity:.4f}",
                    hoverinfo="text",
                    legendgroup=readId,
                    showlegend=showlegend,
                    mode=mode,
                    marker=marker,
                    line_color=color,
                )
            )

    # Change legend names to include the count of best windows the sequence is in.
    for readId, theseScatters in scatters.items():
        for scatter in theseScatters:
            scatter["name"] = f"{readId} ({bestCounts[readId]})"

    data = []

    # Add the scatters for each read in the order found in identity.
    for read in identity:
        data.extend(scatters[read.id])

    return data


def getAllLines(
    centers: list[float], identity: dict[DNARead, list[float]], args: argparse.Namespace
) -> list[go.Scatter]:
    """
    Plot identity lines for all sequences.

    @param centers: The x-axis centers of the genome windows.
    @param identity: A mapping from sequences (that have been matched against the
        reference) to a list of float identities (one for each window).
    @param title: The title of the plot.
    @param args: The command-line arguments.
    """
    data = []

    mode, marker = (
        ("lines+markers", dict(size=args.markerSize))
        if args.markerSize  # i.e., not 0 and not None.
        else ("lines", None)
    )

    for read, identities in identity.items():
        data.append(
            go.Scatter(
                x=centers,
                y=identities,
                mode=mode,
                marker=marker,
                name=read.id,
            )
        )

    return data


def plotPlotly(
    centers: list[float],
    identity: dict[DNARead, list[float]],
    title: str,
    args: argparse.Namespace,
) -> None:
    """
    Make a plotly plot.

    @param centers: The x-axis centers of the genome windows.
    @param identity: A mapping from sequences (that have been matched against the
        reference) to a list of float identities (one for each window).
    @param title: The title of the plot.
    @param args: The command-line arguments.
    """
    axisFontSize: int = 16
    titleFontSize: int = 18

    getData = getBestOnly if args.bestOnly else getAllLines
    data = getData(centers, identity, args)

    xaxis = {
        "title": "Genome location",
        "titlefont": {
            "size": axisFontSize,
        },
    }

    yaxis = {
        "title": "Nucleotide identity fraction",
        "titlefont": {
            "size": axisFontSize,
        },
    }

    layout = go.Layout(
        title={"text": title, "x": 0.5, "xanchor": "center"},
        xaxis=xaxis,
        yaxis=yaxis,
        titlefont={
            "size": titleFontSize,
        },
        hovermode="closest",
    )

    fig = go.Figure(data=data, layout=layout)

    if args.html:
        write_html(fig, args.html)

    if args.out:
        write_image(fig, args.out)


def main() -> None:
    """
    Read the FASTA, compute the windowed identity and then make either a matplotlib
    or plotly plot.
    """
    parser = makeParser()
    args = parser.parse_args()

    if not (args.out or args.html):
        sys.exit(
            "You must specify --out or --html to produce a static image or a "
            "plotly HTML image (or both).",
        )

    reads = tuple(
        DNARead(read.id, read.sequence) for read in parseFASTACommandLineOptions(args)
    )

    idsSeen: set[str] = set()
    for read in reads:
        if read.id in idsSeen:
            # We could be much more helpful here. E.g., compare the sequences, give the
            # index of the read, collect all errors and not fail on just the first.
            sys.exit(
                f"Read with id {read.id} seen more than once in the input.",
            )

    wi = WindowedIdentity(reads)

    reference, starts, identity = wi.getIdentity(
        referenceRegex=args.reference,
        window=args.window,
        minWindow=args.minWindow,
        jump=args.jump,
        includeRegex=args.include,
        sort=args.sort,
        strict=args.strict,
        startOffset=args.startOffset,
        stopOffset=args.stopOffset,
    )

    if args.dryRun:
        nToCompare = len(identity)
        s = "" if nToCompare == 1 else "s"
        print(f"The reference sequence is {reference.id!r}, ", end="")
        print(
            f"to which {nToCompare} sequence{s} would be compared:\n  "
            + "\n  ".join(sorted(read.id for read in identity))
        )
        sys.exit(0)

    # We will plot the identity fractions at the midpoints of the windows,
    # so we need the x-axis coordinates of the window centers.
    half = args.window / 2.0
    centers = [start + half for start in starts]

    title = args.title or (
        f"{args.window} nt (jump {args.jump}){' best' if args.bestOnly else ''} "
        f"windowed identity against {reference.id}"
    )

    if args.out and args.matplotlib:
        # Produce a static image using matplotlib.
        plotMatplotlib(centers, identity, title, args)

    if args.html or (args.out and not args.matplotlib):
        # Produce HTML (and maybe a static image) using plotly.
        plotPlotly(centers, identity, title, args)


if __name__ == "__main__":
    main()



import builtins
from io import BytesIO
from unittest import TestCase
from unittest.mock import mock_open, patch

from dark.fasta_ss import SSFastaReads
from dark.reads import SSAARead


class TestSSFastaReads(TestCase):
    """
    Tests for the L{dark.fasta.SSFastaReads} class.
    """

    def testEmpty(self):
        """
        An empty PDB FASTA file results in an empty iterator.
        """
        data = b""
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = SSFastaReads("x.fasta")
            self.assertEqual([], list(reads))

    def testOddNumberOfRecords(self):
        """
        Trying to parse a PDB FASTA file with an odd number of records must
        raise a ValueError.
        """
        data = b"\n".join([b">seq1", b"REDD", b">str1", b"HH--", b">seq2", b"REAA"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            error = "^Structure file 'x.fasta' has an odd number of records\\.$"
            self.assertRaisesRegex(ValueError, error, list, SSFastaReads("x.fasta"))

    def testUnequalSequenceAndStructureLengths(self):
        """
        Trying to parse a PDB FASTA file that has a sequence whose structure
        is of a different length must raise a ValueError.
        """
        data = b"\n".join(
            [b">seq1", b"REDD", b">str1", b"HH--", b">seq2", b"REAA", b">str2", b"HH"]
        )
        with patch.object(builtins, "open", mock_open(read_data=data)):
            error = (
                "Sequence 'seq2' length \\(4\\) is not equal to structure "
                "'str2' length \\(2\\) in input file 'x\\.fasta'\\.$"
            )
            self.assertRaisesRegex(ValueError, error, list, SSFastaReads("x.fasta"))

    def testOneRead(self):
        """
        A PDB FASTA file with one read must be read properly.
        """
        data = b"\n".join([b">seq1", b"REDD", b">str1", b"HH--"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(SSFastaReads("x.fasta"))
            self.assertEqual([SSAARead("seq1", "REDD", "HH--")], reads)

    def testNoQuality(self):
        """
        A PDB FASTA file read must not have any quality information.
        """
        data = b"\n".join([b">seq1", b"REDD", b">str1", b"HH--"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(SSFastaReads("x.fasta"))
            self.assertIs(None, reads[0].quality)

    def testTwoReads(self):
        """
        A PDB FASTA file with two reads must be read properly and its
        sequences must be returned in the correct order.
        """
        data = b"\n".join(
            [b">seq1", b"REDD", b">str1", b"HH--", b">seq2", b"REAA", b">str2", b"HHEE"]
        )
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(SSFastaReads("x.fasta"))
            self.assertEqual(2, len(reads))
            self.assertEqual(
                [SSAARead("seq1", "REDD", "HH--"), SSAARead("seq2", "REAA", "HHEE")],
                reads,
            )

    def testTypeDefaultsToSSAARead(self):
        """
        A PDB FASTA file whose type is not specified must result in reads that
        are instances of SSAARead.
        """
        data = b"\n".join([b">seq1", b"REDD", b">str1", b"HH--"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(SSFastaReads("x.fasta"))
            self.assertTrue(isinstance(reads[0], SSAARead))

    def testReadClass(self):
        """
        A PDB FASTA file whose read class is something other than SSAARead must
        result in reads that are instances of that class.
        """

        class ReadClass:
            def __init__(self, id, sequence, structure):
                pass

        data = b"\n".join([b">seq1", b"RRRR", b">str1", b"HHHH"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(SSFastaReads("x.fasta", readClass=ReadClass))
            self.assertTrue(isinstance(reads[0], ReadClass))

    def testConvertLowerToUpperCaseIfSpecified(self):
        """
        A read sequence and structure must be converted from lower to upper
        case if requested.
        """
        data = b"\n".join([b">seq1", b"rrrff", b">str1", b"hheeh"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(SSFastaReads("x.fasta", upperCase=True))
            self.assertEqual([SSAARead("seq1", "RRRFF", "HHEEH")], reads)

    def testDontConvertLowerToUpperCaseIfNotSpecified(self):
        """
        A read sequence and its structure must not be converted from lower to
        upper case if the conversion is not requested.
        """
        data = b"\n".join([b">seq1", b"rrFF", b">str1", b"HHee"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(SSFastaReads("x.fasta"))
            self.assertEqual([SSAARead("seq1", "rrFF", "HHee")], reads)

    def testTwoFiles(self):
        """
        It must be possible to read from two FASTA files.
        """

        class SideEffect:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("file1.fasta", filename)
                    self.count += 1
                    return BytesIO(b">id1\nACTG\n>id1\nhhhh\n")
                elif self.count == 1:
                    self.test.assertEqual("file2.fasta", filename)
                    self.count += 1
                    return BytesIO(b">id2\nCAGT\n>id2\neeee\n")
                else:
                    self.test.fail("We are only supposed to be called twice!")

        sideEffect = SideEffect(self)
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect.sideEffect
            reads = SSFastaReads(["file1.fasta", "file2.fasta"])
            self.assertEqual(
                [
                    SSAARead("id1", "ACTG", "hhhh"),
                    SSAARead("id2", "CAGT", "eeee"),
                ],
                list(reads),
            )



#!/usr/bin/env python

"""
Read a FASTA or FASTQ file (or read stdin) and print a line for each sequence,
with the length of the sequence followed by its name.
"""

import argparse

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


def main():
    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on stdin, write the sequence lengths and ids to stdout."
        ),
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    for read in reads:
        print(len(read), read.id)


if __name__ == "__main__":
    main()



from dark.reads import DNARead


def findPrimer(primer: str, seq: str) -> list[int]:
    """
    Look for a primer sequence.

    @param primer: A C{str} primer sequence.
    @param seq: A C{str} sequence.

    @return: A C{list} of zero-based offsets into the sequence at which the
        primer can be found. If no instances are found, return an empty
        C{list}.
    """
    offsets = []
    seq = seq.upper()
    primer = primer.upper()
    primerLen = len(primer)
    discarded = 0
    offset = seq.find(primer)

    while offset > -1:
        offsets.append(discarded + offset)
        seq = seq[offset + primerLen :]
        discarded += offset + primerLen
        offset = seq.find(primer)

    return offsets


def findPrimerBidi(primer: str, seq: str) -> tuple[list[int], list[int]]:
    """
    Look for a primer in a sequence and its reverse complement.

    @param primer: A C{str} primer sequence.
    @param seq: A C{str} sequence.

    @return: A C{tuple} of two lists. The first contains (zero-based)
        ascending offsets into the sequence at which the primer can be found.
        The second is a similar list ascending offsets into the original
        sequence where the primer matches the reverse complemented of the
        sequence. If no instances are found, the corresponding list in the
        returned tuple must be empty.
    """
    # Note that we reverse complement the primer to find the reverse matches. This is
    # much simpler than reverse complementing the sequence because it allows us to use
    # findPrimer and to deal with overlapping matches correctly.
    primerRC = DNARead("id", primer).reverseComplement().sequence
    forward = findPrimer(primer, seq)
    reverse = findPrimer(primerRC, seq)
    return forward, reverse


def findPrimerBidiLimits(primer, seq):
    """
    Report the extreme (inner) offsets of primer in a sequence and its
    reverse complement.

    @param primer: A C{str} primer sequence.
    @param seq: A BioPython C{Bio.Seq} sequence.

    @return: A C{tuple} of two C{int} offsets. The first is a (zero-based)
        offset into the sequence that is beyond the first instance (if any)
        of the primer. The second is an offset into the original sequence of
        the beginning of the last instance of the primer in the reverse
        complemented sequence.

        In other words, if you wanted to chop all instances of a primer
        out of a sequence from the start and the end (when reverse
        complemented) you'd call this function and do something like this:

          start, end = findPrimerBidiLimits(primer, seq)
          seq = seq[start:end]
    """
    forward, reverse = findPrimerBidi(primer, seq)
    if forward:
        start = forward[-1] + len(primer)
        end = len(seq)
        for offset in reverse:
            if offset >= start:
                end = offset
                break
    else:
        start = 0
        end = reverse[0] if reverse else len(seq)

    return start, end



#!/usr/bin/env python

"""
Given a set of FASTA sequence identifiers from sys.argv and/or in a file, read
FASTA from stdin, and print FASTA to stdout for the given sequence ids.
"""

import argparse
import sys

from dark.fasta import FastaReads
from dark.reads import Reads


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Extract a subset of FASTA reads by id",
        epilog=(
            "Given a set of FASTA sequence identifiers from sys.argv "
            "or in a file, read FASTA from stdin, and print FASTA to stdout "
            "for the given sequence ids."
        ),
    )

    parser.add_argument("ids", default=None, nargs="*", help="Wanted read ids.")

    parser.add_argument("--readIdFile", default=None, help="A file of wanted read ids.")

    args = parser.parse_args()
    wanted = set()

    if args.ids:
        wanted.update(args.ids)

    if args.readIdFile is not None:
        with open(args.readIdFile) as fp:
            for line in fp:
                wanted.add(line[:-1])

    found = []

    if wanted:
        for seq in FastaReads(sys.stdin):
            if seq.id in wanted:
                wanted.remove(seq.id)
                found.append(seq)

        if found:
            Reads(found).save(sys.stdout)

        print("Found %d sequences." % len(found), file=sys.stderr)

        if wanted:
            print(
                "WARNING: %d sequence%s not found: %s"
                % (len(wanted), "" if len(wanted) == 1 else "s were", " ".join(wanted)),
                file=sys.stderr,
            )
    else:
        # No wanted ids were given.
        parser.print_help(sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



import multiprocessing
import re
import sys
from os.path import join
from shutil import rmtree
from subprocess import CalledProcessError
from tempfile import mkdtemp
from typing import Optional, Tuple

import edlib

from dark.dna import AMBIGUOUS
from dark.fasta import FastaReads
from dark.process import Executor
from dark.reads import Reads

MAFFT_DEFAULT_ARGS = "--globalpair --maxiterate 1000 --preservecase"
NEEDLE_DEFAULT_ARGS = "auto"

EDLIB_AMBIGUOUS: Tuple[Tuple[str, ...], ...] = tuple(
    [
        tuple(sorted((nt, code)))
        for nt, ambiguities in AMBIGUOUS.items()
        for code in ambiguities
        if len(ambiguities) > 1
    ]
)


def mafft(
    reads: Reads,
    verbose: bool = False,
    options: Optional[str] = None,
    threads: Optional[int] = None,
    executor: Optional[Executor] = None,
    dryRun: bool = False,
) -> Reads:
    """
    Run a MAFFT alignment and return the sequences.

    @param reads: An iterable of multiple reads.
    @param verbose: If C{True} print progress info to sys.stderr.
    @param options: A C{str} of options to pass to mafft.
    @param executor: An C{Executor} instance. If C{None}, a new executor
        will be created and used.
    @param dryRun: If C{True}, do not execute commands, just let the executor
        log them.
    @return: A C{Reads} instance with the aligned sequences.
    """
    tempdir = mkdtemp()

    infile = join(tempdir, "input.fasta")
    out = join(tempdir, "result.fasta")

    Reads(reads).save(infile)

    if verbose:
        print("Running mafft.", file=sys.stderr)

    executor = executor or Executor(dryRun=dryRun)

    executor.execute(
        "mafft %s %s '%s' > '%s'"
        % (
            ("" if threads is None else "--thread %d" % threads),
            options or "",
            infile,
            out,
        ),
        dryRun=dryRun,
    )

    # Use 'list' in the following to force reading the FASTA from disk.
    result = Reads([] if dryRun else list(FastaReads(out)))
    rmtree(tempdir)

    return result


def needle(
    reads: Reads,
    verbose: bool = False,
    options: Optional[str] = None,
    executor: Optional[Executor] = None,
    dryRun: bool = False,
) -> Reads:
    """
    Run a Needleman-Wunsch alignment and return the two sequences.

    @param reads: An iterable of two reads.
    @param verbose: If C{True} print progress info to sys.stderr.
    @param options: Additional options to pass to needle.
    @param executor: An C{Executor} instance. If C{None}, a new executor
        will be created and used.
    @param dryRun: If C{True}, do not execute commands, just let the executor
        log them.
    @return: A C{Reads} instance with the two aligned sequences.
    """
    tempdir = mkdtemp()

    file1 = join(tempdir, "file1.fasta")
    with open(file1, "w") as fp:
        print(reads[0].toString("fasta"), end="", file=fp)

    file2 = join(tempdir, "file2.fasta")
    with open(file2, "w") as fp:
        print(reads[1].toString("fasta"), end="", file=fp)

    out = join(tempdir, "result.fasta")

    def useStderr(e):
        return "Sequences too big. Try 'stretcher'" not in e.stderr

    executor = executor or Executor(dryRun=dryRun)

    if verbose:
        print("Running needle.", file=sys.stderr)
    try:
        executor.execute(
            "needle -asequence '%s' -bsequence '%s' %s "
            "-outfile '%s' -aformat fasta" % (file1, file2, options or "", out),
            dryRun=dryRun,
            useStderr=useStderr,
        )
    except CalledProcessError as e:
        if useStderr(e):
            raise
        else:
            if verbose:
                print(
                    "Sequences too long for needle. Falling back to "
                    "stretcher. Be patient!",
                    file=sys.stderr,
                )
            executor.execute(
                "stretcher -asequence '%s' -bsequence '%s' "
                "-auto "
                "-outfile '%s' -aformat fasta" % (file1, file2, out),
                dryRun=dryRun,
            )

    # Use 'list' in the following to force reading the FASTA from disk.
    result = Reads([] if dryRun else list(FastaReads(out)))
    rmtree(tempdir)

    return result


def removeFirstUnnecessaryGaps(
    seq1: str, seq2: str, gapSymbol: str = "-"
) -> Tuple[bool, str, str]:
    """
    Find and remove the first set of gaps in two sequences that can be removed
    without increasing the difference between the strings.

    @param seq1: A C{str} sequence string.
    @param seq2: A C{str} sequence string.
    @param gapSymbol: A C{str} 1-character symbol to use for gaps.
    @return: A 3-C{tuple} with either C{True} and two new C{str} sequences or
        C{False} and two empty strings if no removable gaps were found. The
        C{True} and C{False} first elements are just there to keep mypy happy,
        seeing as I couldn't figure out how to return (None, None) without
        getting warnings.
    """
    assert len(seq1) == len(seq2)
    for start in range(len(seq1)):
        if seq1[start] == gapSymbol:
            if seq2[start] == gapSymbol:
                raise ValueError(
                    f"Sequences {seq1!r} and {seq2!r} both have "
                    f"a gap ({gapSymbol!r}) at offset {start}!"
                )
            first, second = seq1, seq2
            break
        elif seq2[start] == gapSymbol:
            assert seq1[start] != gapSymbol
            first, second = seq2, seq1
            break
    else:
        # Did not find any gaps.
        return False, "", ""

    excessGapCount = 1
    for end in range(start + 1, len(first)):
        if first[end] == gapSymbol:
            assert second[end] != gapSymbol
            excessGapCount += 1
        elif second[end] == gapSymbol:
            excessGapCount -= 1
            if excessGapCount == 0:
                break
    else:
        # We did not get down to zero gaps. This will never happen if we
        # were originally called as a result of a call to edlibAlign
        # (below) and it only calls removeUnnecessaryGaps (also below) when
        # both original sequences have a different length from their
        # aligned versions. In that case there must be at least one gap in
        # both sequences. But for now this is not considered an error
        # because we may not have been called by removeUnnecessaryGaps.
        return False, "", ""

    subseq1 = seq1[start : end + 1]
    subseq1noGaps = subseq1.replace(gapSymbol, "")
    subseq2 = seq2[start : end + 1]
    subseq2noGaps = subseq2.replace(gapSymbol, "")

    diffsWithGaps = sum(a != b for (a, b) in zip(subseq1, subseq2))
    diffsWithoutGaps = sum(a != b for (a, b) in zip(subseq1noGaps, subseq2noGaps))

    if diffsWithoutGaps <= diffsWithGaps:
        # The subsequences match at least as well without gaps, so replace
        # the gapped region in each sequence with its ungapped version.
        return (
            True,
            seq1[:start] + subseq1noGaps + seq1[end + 1 :],
            seq2[:start] + subseq2noGaps + seq2[end + 1 :],
        )
    else:
        return False, "", ""


def removeUnnecessaryGaps(
    seq1: str, seq2: str, gapSymbol: str = "-"
) -> Tuple[str, str]:
    """
    Find and remove all local sets of gaps in two sequences that can be removed
    without increasing the difference between the strings.

    @param seq1: A C{str} sequence string.
    @param seq2: A C{str} sequence string.
    @param gapSymbol: A C{str} 1-character symbol to use for gaps.
    @return: A 2-C{tuple} of C{str} sequences with removable gaps eliminated.
    """
    while True:
        foundGaps, new1, new2 = removeFirstUnnecessaryGaps(seq1, seq2, gapSymbol)
        if foundGaps:
            seq1, seq2 = new1, new2
        else:
            return seq1, seq2


# Note that this has a fixed gap char ("-"). The regex matches a non-empty sequence of
# gaps (via "-+") followed by a non-empty sequence of non-gaps (i.e., nucleotides) (via
# "[^-]+"), and collects both of the matching parts into groups named 'gaps' and 'nt'.
EDLIB_GAP_RE = re.compile("(?P<gaps>-+)(?P<nt>[^-]+)")


def _canonicalizeFirstGap(seq1: str, seq2: str, pos: int = 0) -> Tuple[int, str, str]:
    """
    Canonicalize a first gap (starting from 'pos'), if any.

    @param seq1: A C{str} sequence string.
    @param seq2: A C{str} sequence string.
    @return: A 3-C{tuple} of an int to indicate where we are up to in seq2
        (with -1 to indicate that no canonicalizations were possible)
        and two C{str} sequences with the first gap canonicalized (if possible).
    """

    for match in EDLIB_GAP_RE.finditer(seq2, pos=pos):
        gaps_start, gaps_end = match.span("gaps")
        nt_start, nt_end = match.span("nt")
        nt_len = nt_end - nt_start

        for nt_prefix_len in range(nt_len, 0, -1):
            prefix = seq2[nt_start : nt_start + nt_prefix_len]
            seq1_target = seq1[gaps_start : gaps_start + nt_prefix_len]
            if prefix == seq1_target:
                seq2 = (
                    seq2[:gaps_start]
                    + prefix
                    + "-" * (gaps_end - gaps_start)
                    + seq2[nt_start + nt_prefix_len : nt_end]
                    + seq2[nt_end:]
                )

                return gaps_start + nt_prefix_len, seq1, seq2

    # No canonicalization could be found.
    return -1, seq1, seq2


def canonicalizeAllGaps(seq1: str, seq2: str) -> Tuple[str, str]:
    """
    Canonicalize all gaps by shifting nucleotides to the right of gaps.

    @param seq1: A C{str} sequence string.
    @param seq2: A C{str} sequence string.
    @return: A 2-C{tuple} of C{str} sequences with gaps merged.
    """

    # We canonicalize by moving the nucleotides to the right (i.e., the gaps to the
    # left) because it undoes the greedy matching of edlib which matches as early as it
    # can. In many cases (with closely matching sequences where there is a gap) you
    # would get a better match (fewer gaps) if you waited. You can see an extreme
    # example of this in the final tests (e.g., testLongExample) in
    # ../../tests/test_aligners.py. That test acts on an actual output of edlib. This
    # approach to canonicalization gives you (at least in the small number of cases I
    # have looked at) the same result you get from MAFFT. MAFFT gets it "right" because
    # it has a penalty for opening a new gap, so it is disinclined to make that decision
    # if there is something more parsimonious. That also makes it much slower because it
    # has to examine many (alignment) paths. edlib just zips along doing something very
    # simple (greedy matching). Fortunately it's so basic that it's easy to improve, at
    # least in these obvious cases of just shifting the nucletotides to the right (if
    # they match at the end of the gap edlib has opened).
    #
    # The action takes place in _canonicalizeFirstGap (above), but I'm describing it
    # here where the initial and final sequence reversals happen. In between we
    # repeatedly call _canonicalizeFirstGap until it cannot find anymore gaps to
    # canonicalize.

    pos = 0
    seq1 = seq1[::-1]
    seq2 = seq2[::-1]

    while pos != -1:
        pos, seq1, seq2 = _canonicalizeFirstGap(seq1, seq2, pos)

    return seq1[::-1], seq2[::-1]


def edlibAlign(
    reads: Reads,
    gapSymbol: str = "-",
    minimizeGaps: bool = True,
    onlyTwoSequences: bool = True,
    matchAmbiguous: bool = True,
    canonicalizeGaps: bool = True,
) -> Reads:
    """
    Run an edlib alignment and return the sequences.

    @param reads: An iterable of at least two reads. The reads must not contain
        C{gapSymbol}.
    @param gapSymbol: A C{str} 1-character symbol to use for gaps.
    @param minimizeGaps: Post-process the edlib output to remove unnecessary gaps.
    @param onlyTwoSequences: Ensure that C{reads} only has two reads.
    @param matchAmbiguous: If C{True}, count ambiguous nucleotides that are
        possibly correct as actually being correct. Otherwise, we are strict
        and nucleotide codes only match themselves.
    @param canonicalizeGaps: Post-process the edlib output so that nucleotides
        that could have been at the end of a gap (instead of in the middle of
        a gap, thereby creating two gaps) are moved to the gap end. This results
        in unecessary gaps being merged.

        For example, in this alignment:

            ACTCTGACGTC
            ACT--G---TC

        the G in the second row could equivalently have been placed at the end of
        a single gap:

            ACTCTGACGTC
            ACT-----GTC

        The single gap is more akin to what MAFFT does because it is reluctant to
        open a new gap. edlib doesn't work in that way, it greedily aligns whatever
        it can as it works its way through the two sequences.

        Here's an example where gaps aren't merged but matching nucleotides are
        moved to the end of the gap. This:

            ACTCGACGTC
            ACTCG---TC

        becomes this:

            ACTCGACGTC
            ACT---CGTC

        Here is a more extreme example. This alignment with 13 gaps:

            ATTAAAGGTTTATACCTTCCCAGGTAACAAACCAACTTTCGATCTCTTGTAGATCTGTTCTCTAAACGAAC
            A-----G----AT-C-T-----G-T-----------T--C--TCT-----A-A----------A--CGAAC

        is equivalent (in terms of matched nucleotides) to this alignment with 1 gap:

            ATTAAAGGTTTATACCTTCCCAGGTAACAAACCAACTTTCGATCTCTTGTAGATCTGTTCTCTAAACGAAC
            --------------------------------------------------AGATCTGTTCTCTAAACGAAC

    @raise ValueError: If C{onlyTwoSequences} is C{True} and there are more
       than two reads passed or if the gap symbol is already present in either
       of the reads.
    @return: A C{Reads} instance with the aligned sequences.
    """
    # Align the first two sequences.
    r1, r2, *rest = list(reads)

    # Complain if there were more than two sequences and we were told to be
    # strict.
    if onlyTwoSequences and len(rest):
        raise ValueError(f"Passed {len(rest)} unexpected extra sequences.")

    for read in r1, r2:
        if gapSymbol in read.sequence:
            raise ValueError(
                f"Sequence {read.id!r} contains one or more gap "
                f"characters {gapSymbol!r}."
            )

    alignment = edlib.getNiceAlignment(
        edlib.align(
            r1.sequence,
            r2.sequence,
            mode="NW",
            task="path",
            additionalEqualities=EDLIB_AMBIGUOUS if matchAmbiguous else None,
        ),
        r1.sequence,
        r2.sequence,
        gapSymbol=gapSymbol,
    )

    seq1, seq2 = alignment["query_aligned"], alignment["target_aligned"]

    # Try to remove unneeded gaps if requested. This is only possible if
    # the length of both sequences has changed.
    if minimizeGaps and len(seq1) != len(r1) and len(seq2) != len(r2):
        seq1, seq2 = removeUnnecessaryGaps(seq1, seq2, gapSymbol)

    if canonicalizeGaps:
        # Note that we don't pass the gap symbol here. edlib uses "-" and we can't
        # (and don't want/need to) change that.
        seq1, seq2 = canonicalizeAllGaps(seq1, seq2)

    return Reads([r1.__class__(r1.id, seq1), r2.__class__(r2.id, seq2)])


def align(
    reads: Reads,
    aligner: str,
    alignerOptions: str | None = None,
    verbose: bool = False,
    threads: int = multiprocessing.cpu_count(),
) -> Reads:
    """
    Align two reads.
    """
    if aligner == "mafft":
        options = MAFFT_DEFAULT_ARGS if alignerOptions is None else alignerOptions
        reads = mafft(reads, verbose, options=options, threads=threads)
    elif aligner == "needle":
        options = NEEDLE_DEFAULT_ARGS if alignerOptions is None else alignerOptions
        reads = needle(reads, verbose, options=options)
    else:
        assert aligner == "edlib"
        reads = edlibAlign(reads)

    return reads



#!/usr/bin/env python

"""
Read in BLAST records and print a count of how many sequences
each read matches with, followed by the name of the read.
"""

import sys

from dark.blast.alignments import BlastReadsAlignments
from dark.fasta import FastaReads

if __name__ == "__main__":
    if len(sys.argv) < 3:
        print(
            "Usage: %s reads.fasta BLAST-files.json..." % sys.argv[0], file=sys.stderr
        )
    else:
        readsFile = sys.argv[1]
        jsonFiles = sys.argv[2:]
        reads = FastaReads(readsFile)
        readsAlignments = BlastReadsAlignments(readsFile, jsonFiles)
        for readAlignments in readsAlignments:
            print(len(readAlignments.alignments), readAlignments.read.id)



import builtins
import bz2
import gzip
from collections import Counter
from io import BytesIO, StringIO
from unittest import TestCase
from unittest.mock import mock_open, patch

from dark.utils import (
    asHandle,
    baseCountsToStr,
    countPrint,
    entropy2,
    intsToIntervals,
    intsToStringIntervals,
    matchOffset,
    median,
    nucleotidesToStr,
    numericallySortFilenames,
    parseRangeExpression,
    parseRangeString,
    pct,
    take,
)


class TestNumericallySortFilenames(TestCase):
    """
    Test the numericallySortFilenames function.
    """

    def testNoNames(self):
        """
        An empty list must be returned when an empty list is given.
        """
        self.assertEqual([], numericallySortFilenames([]))

    def testOneNonNumericName(self):
        """
        A list with a single non-numeric name should result in that same
        name being returned.
        """
        self.assertEqual(["hey"], numericallySortFilenames(["hey"]))

    def testOneNumericName(self):
        """
        A list with a single numeric name should result in that same
        name being returned.
        """
        self.assertEqual(["3.json"], numericallySortFilenames(["3.json"]))

    def testSeveralNames(self):
        """
        A list with several numeric names should result in a correctly
        sorted list of names being returned.
        """
        self.assertEqual(
            ["1.json", "2.json", "3.json"],
            numericallySortFilenames(["3.json", "1.json", "2.json"]),
        )

    def testSeveralNamesWithUnequalPrefixLengths(self):
        """
        A list with several numeric names whose numeric prefixes differ
        in length should result in a correctly sorted list of names being
        returned.
        """
        self.assertEqual(
            ["2.json", "3.json", "21.json", "35.json", "250.json"],
            numericallySortFilenames(
                ["3.json", "21.json", "35.json", "250.json", "2.json"]
            ),
        )

    def testBasename(self):
        """
        Sorting must be according to file basename.
        """
        self.assertEqual(
            [
                "../output/2.json",
                "../output/3.json",
                "../output/21.json",
                "../output/35.json",
                "../output/250.json",
            ],
            numericallySortFilenames(
                [
                    "../output/3.json",
                    "../output/21.json",
                    "../output/35.json",
                    "../output/250.json",
                    "../output/2.json",
                ]
            ),
        )


class TestMedian(TestCase):
    """
    Tests for the median function.
    """

    def testEmptyArgRaises(self):
        """
        An empty list must cause median to raise ValueError.
        """
        error = "^arg is an empty sequence$"
        self.assertRaisesRegex(ValueError, error, median, [])

    def testMedianOfOne(self):
        """
        The median function must work on a list of length one.
        """
        self.assertEqual(3, median([3]))

    def testMedianOfTwo(self):
        """
        The median function must work on a list of length two.
        """
        self.assertEqual(4.5, median([3.1, 5.9]))

    def testMedianOfThree(self):
        """
        The median function must work on a list of length threee.
        """
        self.assertEqual(5.9, median([3.1, 7.6, 5.9]))

    def testMedianOfFour(self):
        """
        The median function must work on a list of length four.
        """
        self.assertEqual(4.5, median([3.1, 1.3, 7.6, 5.9]))

    def testMedianOfFive(self):
        """
        The median function must work on a list of length five.
        """
        self.assertEqual(5.9, median([3.1, 1.3, 7.6, 9.9, 5.9]))


class TestAsHandle(TestCase):
    """
    Test the asHandle function
    """

    def testOpenFile(self):
        """
        When an open file pointer is passed to asHandle, that same file
        pointer must be returned.
        """
        with patch.object(builtins, "open", mock_open()):
            fp = open("file")
            with asHandle(fp) as newfp:
                self.assertIs(fp, newfp)

    def testStr(self):
        """
        When a string filename is passed to asHandle, it must be possible to
        read the correct data from the fp that is returned.
        """
        mockOpener = mock_open(read_data="xxx")
        with patch.object(builtins, "open", mockOpener):
            with asHandle("file") as fp:
                self.assertEqual("xxx", fp.read())

    def testBZ2(self):
        """
        When a string '*.bz2' filename is passed to asHandle, it must be
        possible to read the correct data from the fp that is returned.
        """
        result = BytesIO(b"xxx")

        with patch.object(bz2, "BZ2File") as mockMethod:
            mockMethod.return_value = result
            with asHandle("file.bz2") as fp:
                self.assertEqual("xxx", fp.read())

    def testGzip(self):
        """
        When a string '*.gz' filename is passed to asHandle, it must be
        possible to read the correct data from the fp that is returned.
        """
        result = BytesIO(b"xxx")

        with patch.object(gzip, "GzipFile") as mockMethod:
            mockMethod.return_value = result
            with asHandle("file.gz") as fp:
                self.assertEqual("xxx", fp.read())


class TestParseRangeString(TestCase):
    """
    Check that the parseRangeString function works as expected.
    """

    def testEmptyString(self):
        """
        An empty string must produce an empty set of indices.
        """
        error = "^Illegal range ''. Ranges must single numbers or number-number\\.$"
        self.assertRaisesRegex(ValueError, error, parseRangeString, "")

    def testSingleNumber(self):
        """
        A single number must result in the expected set.
        """
        self.assertEqual({6}, parseRangeString("6"))

    def testSingleNumberSpaceBefore(self):
        """
        A single number preceeded by whitespace must result in the expected
        set.
        """
        self.assertEqual({6}, parseRangeString("  6"))

    def testSingleNumberSpaceAfter(self):
        """
        A single number followed by whitespace must result in the expected
        set.
        """
        self.assertEqual({6}, parseRangeString("6  "))

    def testSingleNumberSpaceBeforeAndAfter(self):
        """
        A single number preceeded and followed by whitespace must result in
        the expected set.
        """
        self.assertEqual({6}, parseRangeString(" 6  "))

    def testSingleRange(self):
        """
        A single range must result in the expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6-10"))

    def testSingleRangeWithSpaceBeforeHyphen(self):
        """
        A single range with a space before the hyphen must result in the
        expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6 -10"))

    def testSingleRangeWithSpaceAfterHyphen(self):
        """
        A single range with a space after the hyphen must result in the
        expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6- 10"))

    def testSingleRangeWithSpaceBeforeAfterHyphen(self):
        """
        A single range with spaces before and after the hyphen must result in
        the expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6 - 10"))

    def testTwoRanges(self):
        """
        Two ranges must result in the expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6-8,9-10"))

    def testTwoOverlappingRanges(self):
        """
        Two overlapping ranges must result in the expected set.
        """
        self.assertEqual({6, 7, 8, 9, 10}, parseRangeString("6-9,7-10"))

    def testTwoRangesAndANumber(self):
        """
        Two ranges and a number must result in the expected set.
        """
        self.assertEqual({6, 7, 8, 10}, parseRangeString("6-8,10"))

    def testTwoRangesAndTwoNumbers(self):
        """
        Two ranges and two numbers must result in the expected set.
        """
        self.assertEqual({4, 6, 7, 8, 9, 10, 11, 12}, parseRangeString("6-8,9,10-12,4"))

    def testZeroConversion(self):
        """
        If we ask for zero conversion, the result must be as expected.
        """
        self.assertEqual(
            {3, 5, 6, 7, 8, 9, 10, 11},
            parseRangeString("6-8,9,10-12,4", convertToZeroBased=True),
        )

    def testAsListPreservesOrderAlreadyOrdered(self):
        """
        When asList is True, the returned value must be a list and the order
        of the numbers in the range must be preserved. Here the passed range
        is already in ascending order.
        """
        result = parseRangeString("6-9,7-10", asList=True)
        self.assertIsInstance(result, list)
        self.assertEqual([6, 7, 8, 9, 10], result)

    def testAsListPreservesOrderDisordered(self):
        """
        When asList is True, the returned value must be a list and the order
        of the numbers in the range must be preserved. Here the passed range
        is not in ascending order (and must be ordered identically in the result).
        """
        result = parseRangeString("7-10,1-3", asList=True)
        self.assertIsInstance(result, list)
        self.assertEqual([7, 8, 9, 10, 1, 2, 3], result)

    def testAsListIgnoresDuplicates(self):
        """
        When asList is True, the returned value must not include duplicates.
        """
        result = parseRangeString("7-10,1-3,8-12,2", asList=True)
        self.assertIsInstance(result, list)
        self.assertEqual([7, 8, 9, 10, 1, 2, 3, 11, 12], result)


class TestParseRangeExpression(TestCase):
    """
    Check that the parseRangeExpression function works as expected.
    """

    def testInvalidExpression(self):
        """
        An invalid string must raise a ValueError.
        """
        error = r"^\($"
        self.assertRaisesRegex(ValueError, error, parseRangeExpression, "(")
        error = r"^hey$"
        self.assertRaisesRegex(ValueError, error, parseRangeExpression, "hey")

    def testEmptyString(self):
        """
        An empty string must produce an empty set.
        """
        self.assertEqual(set(), parseRangeExpression(""))

    def testOneRange(self):
        """
        A simple 3-4 string must produce the expected set.
        """
        self.assertEqual({3, 4}, parseRangeExpression("3-4"))

    def testOneRangeZeroBased(self):
        """
        A simple 3-4 string must produce the expected set when
        convertToZeroBased is True.
        """
        self.assertEqual({2, 3}, parseRangeExpression("3-4", True))

    def testCommas(self):
        """
        A simple 3,4,5 string must produce the expected set.
        """
        self.assertEqual({3, 4, 5}, parseRangeExpression("3,4,5"))

    def testCommasAndRange(self):
        """
        A simple 3,4,5-7 string must produce the expected set.
        """
        self.assertEqual({3, 4, 5, 6, 7}, parseRangeExpression("3,4,5-7"))

    def testTwoRanges(self):
        """
        A simple 3-4,6-8 string must produce the expected set.
        """
        self.assertEqual({3, 4, 6, 7, 8}, parseRangeExpression("3-4,6-8"))

    def testTwoRangesWithSpace(self):
        """
        A simple 3-4, 6-8 string must produce the expected set.
        """
        self.assertEqual({3, 4, 6, 7, 8}, parseRangeExpression("3-4, 6-8"))

    def testUnion(self):
        """
        A union such as 3-4 | 6-8 must produce the expected set.
        """
        self.assertEqual({3, 4, 6, 7, 8}, parseRangeExpression("3-4 | 6-8"))

    def testIntersection(self):
        """
        An intersection such as 3-4 & 4-8 must produce the expected set.
        """
        self.assertEqual({4}, parseRangeExpression("3-4 & 4-8"))

    def testDifferenceNoSpaces(self):
        """
        A difference such as 6-10-7-8 must produce the expected set.
        """
        self.assertEqual({6, 9, 10}, parseRangeExpression("6-10-7-8"))

    def testDifferenceWithSpaces(self):
        """
        A difference such as 6-10 - 7-8 must produce the expected set.
        """
        self.assertEqual({6, 9, 10}, parseRangeExpression("6-10 - 7-8"))

    def testParens(self):
        """
        A difference with parentheses such as '(3-5 | 7-9) & 5-7' must produce
        the expected set.
        """
        self.assertEqual({5, 7}, parseRangeExpression("(3-5 | 7-9) & 5-7"))

    def testDoubleParens(self):
        """
        A difference with two parentheses such as '(3-5 | 7-9) & (5-7 | 9-11)'
        must produce the expected set.
        """
        self.assertEqual({5, 7, 9}, parseRangeExpression("(3-5 | 7-9) & (5-7 | 9-11)"))


class TestStringIO(TestCase):
    """
    Tests for our StringIO class.
    """

    def testInitiallyEmpty(self):
        """
        A StringIO instance must initially be empty.
        """
        self.assertEqual("", StringIO().getvalue())

    def testWriteRead(self):
        """
        It must be possible to write and read to/from a StringIO instance as
        normal.
        """
        s = StringIO()
        s.write("hey")
        self.assertEqual("hey", s.getvalue())

    def testInitializedRead(self):
        """
        It must be possible to read from a StringIO instance that is
        initialized on creation.
        """
        s = StringIO("hey")
        self.assertEqual("hey", s.getvalue())

    def testContextManager(self):
        """
        It must be possible to use a StringIO instance as a context manager.
        """
        with StringIO() as s:
            s.write("hey")
            self.assertEqual("hey", s.getvalue())


class TestBaseCountsToStr(TestCase):
    """
    Test the baseCountsToStr function.
    """

    def testSimple(self):
        """
        A simple example must work as expected.
        """
        counts = Counter()
        counts["A"] += 1
        counts["G"] += 2
        self.assertEqual("A:1 G:2", baseCountsToStr(counts))


class TestNucleotidesToStr(TestCase):
    """
    Test the nucleotidesToStr function.
    """

    def testSimple(self):
        """
        A simple example must work as expected.
        """
        counts1 = Counter()
        counts1["A"] += 1
        counts1["G"] += 2
        counts2 = Counter()
        counts2["C"] += 1
        counts2["T"] += 3
        self.assertEqual(
            "0: A:1 G:2\n7: C:1 T:3",
            nucleotidesToStr(
                {
                    0: counts1,
                    7: counts2,
                }
            ),
        )


class TestCountPrint(TestCase):
    """
    Test the countPrint function and the contained percentage function.
    """

    def testSimple(self):
        """
        A simple example must work as expected.
        """
        count = 2
        len1 = 10
        self.assertEqual("Count is: 2/10 (20.00%)", countPrint("Count is", count, len1))

    def testTwoSequences(self):
        """
        An example involving two different lengths must work as expected.
        """
        count = 2
        len1 = 10
        len2 = 8
        self.assertEqual(
            "Count is: 2/10 (20.00%) of sequence 1, 2/8 (25.00%) of sequence 2",
            countPrint("Count is", count, len1, len2),
        )


class TestPct(TestCase):
    """
    Test the pct function.
    """

    def testZeroNumerator(self):
        """
        The pct function must produce the correct result if the numerator is
        zero.
        """
        self.assertEqual("0/10 (0.000%)", pct(0, 10))

    def testZeroDenominator(self):
        """
        The pct function must produce the correct result if the denominator is
        zero.
        """
        self.assertEqual("0/0 (0.000%)", pct(0, 0))

    def testZeroDenominatorPlaces(self):
        """
        The pct function must produce the correct result if the denominator is
        zero and also respect the passed value of places.
        """
        self.assertEqual("0/0 (0.0%)", pct(0, 0, places=1))

    def testOneHalf(self):
        """
        The pct function must produce the correct result if the numerator is
        one half of the denominator.
        """
        self.assertEqual("5/10 (50.000%)", pct(5, 10))

    def testPlaces(self):
        """
        It must be possible to specify the number of decimal places.
        """
        self.assertEqual("5/10 (50.00%)", pct(5, 10, places=2))

    def testOneSeventh(self):
        """
        The pct function must produce the correct result if the numerator is
        one seventh of the denominator.
        """
        self.assertEqual("2/14 (14.286%)", pct(2, 14))

    def testCommas(self):
        """
        The pct function must be able to produce a result with commas.
        """
        self.assertEqual("1,024/2,048 (50.000%)", pct(1024, 2048, commas=True))

    def testNoDenominatorWithoutCommas(self):
        """
        The pct function must be able to produce a result with no denominator
        when no commas are requested.
        """
        self.assertEqual("1024 (50.000%)", pct(1024, 2048, denom=False))

    def testNoDenominatorWithCommas(self):
        """
        The pct function must be able to produce a result with no denominator
        when commas=True.
        """
        self.assertEqual("1,024 (50.000%)", pct(1024, 2048, commas=True, denom=False))


class TestTake(TestCase):
    """
    Test the take function.
    """

    def testNLessThanOne(self):
        """
        The take function must raise an AssertionError if passed n < 1.
        list.
        """
        self.assertRaises(AssertionError, next, take([], 0))

    def testEmpty(self):
        """
        The take function must return an empty generator if passed an empty
        list.
        """
        self.assertEqual([], list(take([], 4)))

    def testOne(self):
        """
        The take function must return individual items if passed n=1.
        """
        self.assertEqual([[3], [4], [5]], list(take([3, 4, 5], 1)))

    def testNBiggerThanPassedList(self):
        """
        The take function must return all items if passed an n that is bigger
        than the passed list.
        """
        self.assertEqual([[3, 4, 5]], list(take([3, 4, 5], 100)))

    def testThree(self):
        """
        The take function must return three items at a time if passed n=3, and
        also return the extra two items.
        """
        self.assertEqual(
            [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10]], list(take(range(11), 3))
        )


class TestMatchOffset(TestCase):
    """
    Test the matchOffset function.
    """

    def testEmpty(self):
        """
        An empty query must match an empty reference at position 0.
        """
        self.assertEqual(0, matchOffset("", ""))

    def testEqualStrings(self):
        """
        An non-empty reference must match an identical non-empty query at
        position 0.
        """
        self.assertEqual(0, matchOffset("AA", "AA"))

    def testQueryPaddedLeftByOne(self):
        """
        A query that is padded on the left by one space must match at
        position 1.
        """
        self.assertEqual(1, matchOffset("AA", " A"))

    def testQueryPaddedLeftByTwo(self):
        """
        A query that is padded on the left by two spaces must match at
        position 2.
        """
        self.assertEqual(2, matchOffset("AAA", "  A"))

    def testQueryPaddedLeftByTwoReferencePaddedLeftByOne(self):
        """
        A query that is padded on the left by two spaces must match a reference
        that is padded on the left by one space at position 1.
        """
        self.assertEqual(1, matchOffset(" AA", "  A"))

    def testQueryPaddedLeftByFiveReferencePaddedLeftByOneWithGaps(self):
        """
        A query that is padded on the left by five spaces must correctly match
        a reference containing gaps that is padded on the left by one space.
        """
        self.assertEqual(2, matchOffset(" AA--G", "     A"))

    def testQueryPaddedLeftByFiveReferenceOneGapLeftWithGaps(self):
        """
        A query that is padded on the left by five spaces must correctly match
        a reference that starts with a gap and that contains gaps.
        """
        self.assertEqual(2, matchOffset("-AA--G", "     A"))

    def testOmicronPartialInsertionRead(self):
        """
        A query that overlaps part of an insertion in the reference must be
        handled correctly.
        """
        self.assertEqual(
            12,
            matchOffset(
                "TAATTTAGTGCG---------TGATCTCCCTCAGGGTTTTTCGGCTTTAGAAC",
                "             AGCCAGAATGATCTCCCTCAGGGTTTTTCGGCTTT",
            ),
        )


class TestIntsToIntervals(TestCase):
    """
    Test the intsToIntervals function.
    """

    def testEmpty(self):
        "An empty set of numbers must give an empty result."
        self.assertEqual(intsToIntervals([]), [])

    def testOneNumber(self):
        "A set of one number must result in that number."
        self.assertEqual(intsToIntervals({2018}), [(2018, 2018)])

    def testTwoNonConsecutiveNumbers(self):
        """
        A set of two numbers that are not consecutive must result in those numbers.
        """
        self.assertEqual(intsToIntervals({2018, 2020}), [(2018, 2018), (2020, 2020)])

    def testTwoNonConsecutiveNumbersDecreasingOrder(self):
        """
        A set of two numbers that are not consecutive (and which decrease in the
        passed order) must result in those numbers.
        """
        self.assertEqual(intsToIntervals([2020, 2018]), [(2018, 2018), (2020, 2020)])

    def testTwoConsecutiveNumbers(self):
        """
        A set of two numbers that are consecutive must result in those numbers in a
        range.
        """
        self.assertEqual(intsToIntervals({2018, 2019}), [(2018, 2019)])

    def testThreeConsecutiveNumbers(self):
        """
        A set of three numbers that are consecutive must result in those numbers in a
        range.
        """
        self.assertEqual(intsToIntervals({2018, 2019, 2020}), [(2018, 2020)])

    def testTwoConsecutiveIntervals(self):
        """
        Two set of consecutive numbers must result in those numbers in two
        intervals.
        """
        self.assertEqual(
            intsToIntervals({2018, 2019, 2020, 2014, 2015}),
            [(2014, 2015), (2018, 2020)],
        )

    def testOneToAHundred(self):
        """
        1 to 100 must result in "1-100".
        """
        self.assertEqual(intsToIntervals(range(1, 101)), [(1, 100)])

    def testNegativeRange(self):
        """
        A range from one negative number to another must give the expected result.
        """
        self.assertEqual(intsToIntervals({-2018, -2019}), [(-2019, -2018)])

    def testPositiveAndNegativeIntervals(self):
        """
        Two intervals, one negative and one positive must give the expected result.
        """
        self.assertEqual(
            intsToIntervals({-2020, -2018, -2019, 2010, 2011, 2012}),
            [(-2020, -2018), (2010, 2012)],
        )


class TestIntsToStringIntervals(TestCase):
    """
    Test the intsToStringIntervals function.
    """

    def testEmpty(self):
        "An empty set of numbers must give an empty result."
        self.assertEqual(intsToStringIntervals([]), [])

    def testOneNumber(self):
        "A set of one number must result in that number."
        self.assertEqual(intsToStringIntervals({2018}), ["2018"])

    def testTwoNonConsecutiveNumbers(self):
        """
        A set of two numbers that are not consecutive must result in those numbers.
        """
        self.assertEqual(intsToStringIntervals({2018, 2020}), ["2018", "2020"])

    def testTwoNonConsecutiveNumbersDecreasingOrder(self):
        """
        A set of two numbers that are not consecutive (and which decrease in the
        passed order) must result in those numbers.
        """
        self.assertEqual(intsToStringIntervals([2020, 2018]), ["2018", "2020"])

    def testTwoConsecutiveNumbers(self):
        """
        A set of two numbers that are consecutive must result in those numbers in a
        range.
        """
        self.assertEqual(intsToStringIntervals({2018, 2019}), ["2018-2019"])

    def testTwoConsecutiveNumbersAlternateSeparator(self):
        """
        A set of two numbers that are consecutive must result in those numbers in a
        range, separated by the passed separator string.
        """
        self.assertEqual(
            intsToStringIntervals({2018, 2019}, sep=" -> "), ["2018 -> 2019"]
        )

    def testThreeConsecutiveNumbers(self):
        """
        A set of three numbers that are consecutive must result in those numbers in a
        range.
        """
        self.assertEqual(intsToStringIntervals({2018, 2019, 2020}), ["2018-2020"])

    def testTwoConsecutiveIntervals(self):
        """
        Two set of consecutive numbers must result in those numbers in two
        intervals.
        """
        self.assertEqual(
            intsToStringIntervals({2018, 2019, 2020, 2014, 2015}),
            ["2014-2015", "2018-2020"],
        )

    def testOneToAHundred(self):
        """
        1 to 100 must result in "1-100".
        """
        self.assertEqual(intsToStringIntervals(range(1, 101)), ["1-100"])

    def testNegativeRange(self):
        """
        A range from one negative number to another must give the expected result.
        """
        self.assertEqual(
            intsToStringIntervals({-2018, -2019}, sep=" -> "), ["-2019 -> -2018"]
        )

    def testPositiveAndNegativeIntervals(self):
        """
        Two intervals, one negative and one positive must give the expected result.
        """
        self.assertEqual(
            intsToStringIntervals({-2020, -2018, -2019, 2010, 2011, 2012}, sep=" -> "),
            ["-2020 -> -2018", "2010 -> 2012"],
        )

    def testTwoConsecutiveNumbersMinimized(self):
        """
        When minimize=True is passed, a set of two numbers that are consecutive must
        result in those numbers in a simplified range.
        """
        self.assertEqual(intsToStringIntervals({2018, 2019}, minimize=True), ["2018-9"])

    def testTwoConsecutiveNegativeNumbersMinimized(self):
        """
        When minimize=True is passed, a set of two negative numbers that are
        consecutive must result in those numbers in a simplified range.
        """
        self.assertEqual(
            intsToStringIntervals(
                {-2020, -2021, -2018, -2019}, minimize=True, sep=" -> "
            ),
            ["-2021 -> 18"],
        )

    def testThreeConsecutiveNumbersMinimized(self):
        """
        When minimize=True is passed, a set of three numbers that are consecutive must
        result in those numbers in a simplified range.
        """
        self.assertEqual(
            intsToStringIntervals({2018, 2019, 2020}, minimize=True), ["2018-20"]
        )

    def testNegativePositiveRangeIsNotMinimized(self):
        """
        When minimize=True is passed, a set of numbers that are consecutive and
        crossing zero must result in that number range not being minimized.
        """
        self.assertEqual(
            intsToStringIntervals(range(-20, 21), minimize=True, sep=" -> "),
            ["-20 -> 20"],
        )


class TestEntropy2(TestCase):
    """
    Test the entropy2 function.
    """

    def testOneLabel(self):
        self.assertEqual(0.0, entropy2(("a",)))

    def testOneLabelWithCopies(self):
        self.assertEqual(0.0, entropy2(("a", "a", "a")))

    def testTwoLabels(self):
        self.assertEqual(1.0, entropy2(("a", "b")))

    def testTwoLabelsWithCopies(self):
        self.assertEqual(1.0, entropy2(("a", "a", "b", "b")))

    def testFourLabels(self):
        self.assertEqual(2.0, entropy2(("a", "b", "c", "d")))



#!/usr/bin/env python

import sys

from dark.analyze_reads import getPrefixAndSuffix, trimReads
from dark.reads import Reads

if len(sys.argv) != 2:
    sys.exit(f"{sys.argv[0]} takes one FASTA filename argument")

else:
    filename = sys.argv[1]
    with open(filename) as fp:
        prefix, suffix = getPrefixAndSuffix(fp)

    print(f"Prefix length {prefix}, suffix length {suffix}", file=sys.stderr)

    with open(filename) as fp:
        Reads(trimReads(prefix, suffix, fp)).save(sys.stdout)



from unittest import TestCase

from ete4 import Tree

from dark.trees import ete4root


class NewickTreeTester:
    """
    Mixin for testing Newick trees.
    """

    def makeTree(self) -> Tree:
        return Tree(self.newick, parser=1)  # type: ignore[attr-defined]

    def check(self, tree, expected):
        self.assertEqual(  # type: ignore[attr-defined]
            expected, tree.write(parser=1, format_root_node=True)
        )


class Test_Ancestor_MRCA_Unrooted(TestCase):
    """
    Some simple ancestor and MRCA tests on an unrooted tree.
    """

    newick = "(A, B, C);"

    def testSimple(self) -> None:
        tree = Tree(self.newick)
        a_nodes = list(tree.search_nodes(name="A"))
        assert len(a_nodes) == 1
        a_tip = a_nodes[0]
        ancestors = list(a_tip.ancestors(include_root=False))
        assert len(ancestors) == 1
        assert ancestors[0] == tree.root

    def testSimple_Parser1(self) -> None:
        tree = Tree(self.newick, parser=1)
        a_nodes = list(tree.search_nodes(name="A"))
        assert len(a_nodes) == 1
        a_tip = a_nodes[0]
        ancestors = list(a_tip.ancestors(include_root=False))
        assert len(ancestors) == 1
        assert ancestors[0] == tree.root

    def testAncestorsAndMRCAOfTwo(self) -> None:
        tree = Tree(self.newick)
        a_nodes = list(tree.search_nodes(name="A"))
        assert len(a_nodes) == 1
        a_tip = a_nodes[0]

        a_ancestors = list(a_tip.ancestors(include_root=False))
        assert len(a_ancestors) == 1
        assert a_ancestors[0] == tree.root

        b_nodes = list(tree.search_nodes(name="B"))
        assert len(b_nodes) == 1
        b_tip = b_nodes[0]

        b_ancestors = list(b_tip.ancestors(include_root=False))
        assert len(b_ancestors) == 1
        assert b_ancestors[0] == tree.root

        mrca = tree.common_ancestor(a_tip, b_tip)
        assert len(mrca.children) == 3
        assert mrca == tree.root

        included = {a_tip, b_tip}
        for tip in a_tip, b_tip:
            # The include_root is not important. Works either way (confirmed).
            ancestors = tip.ancestors(include_root=False)
            included.update(ancestors)

        excluded = set(mrca.children) - included
        assert len(excluded) == 1
        assert excluded.pop().name == "C"


class Test_Reroot_Rooted_ABC(TestCase, NewickTreeTester):
    r"""
    Test the ete4root function on a rooted tree with three nodes, as below.

       /-A
    --|
      |   /-B
       \-|
          \-C

    """

    newick = "(A, (B, C)BC)ABC;"

    def testRoot_A(self) -> None:
        """
        If we root on the branch leading to A we should get the same tree
        back since this would add a node on that branch and discard the old
        root, resulting in the same topology.
        """
        result = ete4root(self.makeTree(), ("A",))
        self.check(result, "(A,(B,C)BC);")

    def testRoot_B(self) -> None:
        """
        If we root on the branch leading to B we should get the expected result.
        Note that the node with name BC is preserved but it is the node whose
        children are A/C after the re-rooting.
        """
        result = ete4root(self.makeTree(), ("B",))
        self.check(result, "(B,(C,A)BC);")

    def testRoot_C(self) -> None:
        """
        If we root on the branch leading to C we should get the expected result.
        Note that the node with name BC is preserved but it is the node whose
        children are A/C after the re-rooting.
        """
        result = ete4root(self.makeTree(), ("C",))
        self.check(result, "(C,(B,A)BC);")

    def testRoot_BC(self) -> None:
        """
        If we root on the branch leading to B/C we should get the same tree
        back since this would add a node on that branch and discard the old
        root, resulting in the same topology.
        """
        result = ete4root(self.makeTree(), ("B", "C"))
        self.check(result, "((B,C)BC,A);")

    def testDetach_A(self) -> None:
        """
        If we root on the branch leading to A and also ask for the outgroup
        to be detached, we should get back the B/C tree.
        """
        result = ete4root(self.makeTree(), ("A",), detach=True)
        self.check(result, "(B,C)BC:0;")

    def testDetach_BC(self) -> None:
        """
        If we root on the branch leading to B/C and also ask for the outgroup
        to be detached, we should get back the A tree.
        """
        print(self.makeTree())
        result = ete4root(self.makeTree(), ("B", "C"), detach=True)
        self.check(result, "A:0;")


class Test_Reroot_Rooted_ABC_And_Scale(TestCase, NewickTreeTester):
    r"""
    Test the ete4root function (with subsequent scaling) on a rooted tree with
    three nodes, as below.

       /-A
    --|
      |   /-B
       \-|
          \-C

    """

    newick = "(A, (B, C:1)BC)ABC;"

    def testRoot_C(self) -> None:
        """
        If we root on the branch leading to C we should get the expected result,
        including when we scale the distances to the children of the new root.
        Note that the node with name BC is preserved but it is the node whose
        children are A/C after the re-rooting.
        """
        result = ete4root(self.makeTree(), ("C",), scale=0.25)
        # The distances to C and to the A/B node will both be 2 because the
        # original distance (of 1) to the C node is split between the two
        # children of the inserted root node (so 0.5 in each direction) and
        # we pass scale=0.25 which the new distances are divided by (and 0.5
        # / 0.25 = 2).
        # self.check(result, "(C:2,(B:1,A:2)BC:2)ABC:0;")
        self.check(result, "(C:2,(B,A)BC:2);")

        result = ete4root(self.makeTree(), ("C",), scale=0.1)
        self.check(result, "(C:5,(B,A)BC:5);")

        result = ete4root(self.makeTree(), ("C",), scale=2.0)
        self.check(result, "(C:0.25,(B,A)BC:0.25);")


class Test_Reroot_Unrooted_ABC(TestCase, NewickTreeTester):
    r"""
    Test the ete4root function on an unrooted tree with three nodes, as below.

       /-A
      |
    --|--B
      |
       \-C

    """

    newick = "(A, B, C)ABC;"

    def testRoot_A(self) -> None:
        """
        If we root on the branch leading to A we should get the expected result.
        Note that the node with name ABC is preserved but it is the node whose
        children are B/C after the re-rooting.
        """
        result = ete4root(self.makeTree(), ("A",))
        self.check(result, "(A,(B,C)ABC);")

    def testRoot_AB(self) -> None:
        """
        If we root on the branch leading to the MRCA of A/B we should get
        a tree with C on one side and an A/B subtree on the other. Note that
        the node with name ABC is preserved but it is the node whose children
        are B/C after the re-rooting.
        """
        result = ete4root(self.makeTree(), ("A", "B"))
        self.check(result, "(C,(A,B)ABC);")


class Test_Reroot_Unrooted_bigger_tree(TestCase, NewickTreeTester):
    r"""
    Test the ete4root function on the bigger unrooted tree below.

       /-A
      |
      |   /-B
      |--|
    --|  |   /-C
      |   \-|
      |      \-D
      |
      |   /-E
       \-|
         |   /-F
          \-|
             \-G

    """

    newick = "(A, (B, (C, D)), (E, (F, G)));"

    def testRoot_A(self) -> None:
        """
        Re-rooting on the branch that leads to A should work as expected.
        """
        result = ete4root(self.makeTree(), ("A",))
        # result looks like this (you can see it via print(result)).
        r"""
           /-A
          |
          |      /-B
        --|   /-|
          |  |  |   /-C
          |  |   \-|
           \-|      \-D
             |
             |   /-E
              \-|
                |   /-F
                 \-|
                    \-G
        """
        self.check(result, "(A,((B,(C,D)),(E,(F,G))));")

    def testRoot_DG(self) -> None:
        """
        Re-rooting on the MRCA of D and G is the same as rooting on the branch that
        leads to A.
        """
        result = ete4root(self.makeTree(), ("D", "G"))
        # result looks like this (you can see it via print(result)).
        r"""
           /-A
          |
          |      /-B
        --|   /-|
          |  |  |   /-C
          |  |   \-|
           \-|      \-D
             |
             |   /-E
              \-|
                |   /-F
                 \-|
                    \-G
        """
        self.check(result, "(A,((B,(C,D)),(E,(F,G))));")

    def testRoot_EG(self) -> None:
        """
        Re-rooting on the MRCA of E and G should work as expected.
        """
        result = ete4root(self.makeTree(), ("E", "G"))
        # result looks like this (you can see it via print(result)).
        r"""
              /-E
           /-|
          |  |   /-F
          |   \-|
        --|      \-G
          |
          |   /-A
           \-|
             |   /-B
              \-|
                |   /-C
                 \-|
                    \-D
        """
        self.check(result, "((E,(F,G)),(A,(B,(C,D))));")

    def testRoot_CD(self) -> None:
        """
        Re-rooting on the MRCA of C and D should work as expected.
        """
        result = ete4root(self.makeTree(), ("C", "D"))
        # result looks like this (you can see it via print(result)).
        r"""
              /-C
           /-|
          |   \-D
        --|
          |   /-B
           \-|
             |   /-A
              \-|
                |   /-E
                 \-|
                   |   /-F
                    \-|
                       \-G
        """
        self.check(result, "((C,D),(B,(A,(E,(F,G)))));")

    def testRoot_CD_detach(self) -> None:
        """
        Re-rooting on the MRCA of C and D should work as expected when we also then
        detach the C/D subtree.
        """
        result = ete4root(self.makeTree(), ("C", "D"), detach=True)
        # result looks like this (you can see it via print(result)).
        r"""
           /-B
        --|
          |   /-A
           \-|
             |   /-E
              \-|
                |   /-F
                 \-|
                    \-G
        """
        self.check(result, "(B,(A,(E,(F,G)))):0;")



#!/usr/bin/env python

import argparse

from dark.fasta import FastaReads

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=(
            "For a fasta file with sequences, summarize what is "
            "happening at a specific position."
        ),
    )

    parser.add_argument(
        "--fastaFile", required=True, help="The name of the FASTA file to read."
    )

    parser.add_argument(
        "--position",
        required=True,
        type=int,
        help="The (one-based) position to examine.",
    )

    args = parser.parse_args()
    reads = FastaReads(args.fastaFile)
    result = reads.summarizePosition(args.position - 1)
    nReads = result["readCount"]

    print(
        "%d of %d sequences were excluded due to length."
        % (result["excludedCount"], nReads)
    )

    denominator = (nReads - result["excludedCount"]) / 100.0
    for base, count in result["countAtPosition"].items():
        print("%s: Total: %s (%.2f%%)" % (base, count, count / denominator))



from functools import total_ordering


@total_ordering
class HigherIsBetterScore:
    """
    Provide comparison functions for scores where numerically higher values
    are considered better.

    @param score: The numeric score of this HSP.
    """

    def __init__(self, score: float) -> None:
        self.score = score

    def __lt__(self, other: object) -> bool:
        if isinstance(other, HigherIsBetterScore):
            return self.score < other.score
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, HigherIsBetterScore):
            return self.score == other.score
        return NotImplemented

    def betterThan(self, score: float) -> bool:
        """
        Compare this score with another score.

        @param score: A C{float} score.
        @return: A C{bool}, C{True} if this score is the better.
        """
        return self.score > score


@total_ordering
class LowerIsBetterScore:
    """
    Provide comparison functions for scores where numerically lower values
    are considered better.

    @param score: The numeric score of this LSP.
    """

    def __init__(self, score: float) -> None:
        self.score = score

    def __lt__(self, other: object) -> bool:
        if isinstance(other, LowerIsBetterScore):
            return self.score > other.score
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LowerIsBetterScore):
            return self.score == other.score
        return NotImplemented

    def betterThan(self, score: float) -> bool:
        """
        Compare this score with another score.

        @param score: A C{float} score.
        @return: A C{bool}, C{True} if this score is the better.
        """
        return self.score < score



from unittest import TestCase

from Bio.Seq import Seq

from dark.sequence import findPrimer, findPrimerBidi, findPrimerBidiLimits


class TestFindPrimer(TestCase):
    """
    Tests for the dark.sequence.findPrimer function.
    """

    def testNotFound(self):
        """
        If a primer is not found, the empty list must be returned.
        """
        seq = Seq("ACGT")
        self.assertEqual([], findPrimer("BLAH", seq))

    def testFoundAtStart(self):
        """
        If a primer is found at the start of a sequence, a list containing 0
        must be returned.
        """
        seq = Seq("ACGT")
        self.assertEqual([0], findPrimer("AC", seq))

    def testFoundAtEnd(self):
        """
        If a primer is found at the end of a sequence, the correct value
        must be returned.
        """
        seq = Seq("ACGT")
        self.assertEqual([2], findPrimer("GT", seq))

    def testFoundMultiple(self):
        """
        If a primer is found multiple times, the correct value
        must be returned.
        """
        seq = Seq("ACGTACGT")
        self.assertEqual([0, 4], findPrimer("ACG", seq))

    def testOverlapping(self):
        """
        If a primer is present twice but is overlapping, only the first
        instance should be returned.
        """
        seq = Seq("GAAA")
        self.assertEqual([1], findPrimer("AA", seq))


class TestFindPrimerBidi(TestCase):
    """
    Tests for the dark.sequence.findPrimerBidi function.
    """

    def testNotFound(self):
        """
        If a primer is not found, empty lists must be returned.
        """
        seq = Seq("ACGT")
        self.assertEqual(([], []), findPrimerBidi("BLAH", seq))

    def testFoundStartEnd(self):
        """
        If a primer is found in both directions in a sequence (start of
        the forward sequence, end of the reverse complement), the
        correct value must be returned.
        """
        seq = Seq("ACGT")
        self.assertEqual(([0], [2]), findPrimerBidi("AC", seq))

    def testFoundEndStart(self):
        """
        If a primer is found in both directions in a sequence (end of
        the forward sequence, start of the reverse complement), the
        correct value must be returned.
        """
        seq = Seq("ACGT")
        self.assertEqual(([2], [0]), findPrimerBidi("GT", seq))

    def testFoundMultiple(self):
        """
        If a primer is found multiple times, the correct value
        must be returned.
        """
        seq = Seq("ACGTACGT")
        self.assertEqual(([0, 4], [1, 5]), findPrimerBidi("ACG", seq))

    def testOverlappingForwards(self):
        """
        If a primer is present twice forwards but is overlapping, only
        the first instance should be returned.
        """
        seq = Seq("GAAA")
        self.assertEqual(([1], []), findPrimerBidi("AA", seq))

    def testOverlappingBackwards(self):
        """
        If a primer is present twice backwards but is overlapping, only
        the first instance should be returned.
        """
        seq = Seq("GTTT")
        self.assertEqual(([], [1]), findPrimerBidi("AA", seq))


class TestFindPrimerBidiLimits(TestCase):
    """
    Tests for the dark.sequence.findPrimerBidiLimits function.
    """

    def testNotFound(self):
        """
        If a primer is not found, the returned offsets must include
        the whole sequence.
        """
        seq = Seq("ACGT")
        self.assertEqual((0, 4), findPrimerBidiLimits("BLAH", seq))

    def testFoundStartEnd(self):
        """
        If a primer is found in both directions in a sequence (start of
        the forward sequence, end of the reverse complement), the
        correct value must be returned.
        """
        seq = Seq("ACGT")
        self.assertEqual((2, 2), findPrimerBidiLimits("AC", seq))

    def testFoundEndStart(self):
        """
        If a primer is found in both directions in a sequence (end of
        the forward sequence, start of the reverse complement), the
        correct value must be returned.
        """
        seq = Seq("ACGT")
        self.assertEqual((4, 4), findPrimerBidiLimits("GT", seq))

    def testFoundMultiple(self):
        """
        If a primer is found multiple times, the correct value
        must be returned.
        """
        seq = Seq("ACGTACGT")
        self.assertEqual((7, 8), findPrimerBidiLimits("ACG", seq))

    def testOverlappingForwards(self):
        """
        If a primer is present twice forwards but is overlapping, only
        the first instance should be returned.
        """
        seq = Seq("GAAA")
        self.assertEqual((3, 4), findPrimerBidiLimits("AA", seq))

    def testOverlappingBackwards(self):
        """
        If a primer is present twice backwards but is overlapping, only
        the first instance should be returned.
        """
        seq = Seq("GTTT")
        self.assertEqual((0, 1), findPrimerBidiLimits("AA", seq))

    def testLonger(self):
        """
        Test a longer sequence.
        """
        As = "AAAAAAAAAA"
        Gs = "GGGGGGGGGG"
        seq = Seq(As + Gs + 2 * As)
        self.assertEqual((20, 40), findPrimerBidiLimits(Gs, seq))
